# sage.doctest: needs sage.libs.pari
"""
Set and subsets of prime numbers

AUTHORS:

 - William Stein (2005): original version
 - Florent Hivert (2009-11): adapted to the category framework
 - Xavier Caruso (2025-10): implement congruence conditions
"""

# ****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#                     2009 Florent Hivert <Florent.Hivert@univ-rouen.fr>
#                     2025 Xavier Caruso <xavier@caruso.ovh>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.rings.semirings.non_negative_integer_semiring import NN
from sage.rings.integer_ring import ZZ
from sage.rings.infinity import infinity
from .set import Set_generic
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.categories.infinite_enumerated_sets import InfiniteEnumeratedSets
from sage.arith.misc import euler_phi
from sage.structure.unique_representation import UniqueRepresentation
from sage.categories.sets_cat import EmptySetError


def _repr_items(items, left=4, right=2):
    r"""
    Return a string representation of the items in ``items``
    with possible ellipsis.

    INPUT:

    - ``items`` -- a list

    - ``left`` -- an integer (default: ``4``), the maximum
      number of items listed at the beginning

    - ``right`` -- an integer (default: ``2``), the maximum
      number of items listed at the end

    EXAMPLES::

        sage: from sage.sets.primes import _repr_items
        sage: _repr_items(range(5))
        '0, 1, 2, 3, 4'
        sage: _repr_items(range(10))
        '0, 1, 2, 3, ..., 8, 9'
        sage: _repr_items(range(10), left=3, right=3)
        '0, 1, 2, ..., 7, 8, 9'
    """
    if len(items) <= left + right + 1:
        s = [str(item) for item in items]
    else:
        s = [str(item) for item in items[:left]]
        s += ["..."]
        s += [str(item) for item in items[-right:]]
    return ", ".join(s)


class Primes(Set_generic, UniqueRepresentation):
    r"""
    The set of prime numbers and some of its subsets.

    EXAMPLES:

    The set of all primes numbers::

        sage: P = Primes(); P
        Set of all prime numbers: 2, 3, 5, 7, ...

    The arguments ``modulus`` and ``classes`` allows for constructing
    subsets given by congruence conditions::

        sage: Primes(modulus=4)
        Set of prime numbers congruent to 1 modulo 4: 5, 13, 17, 29, ...

    By default the congruence class `1` is selected, but we can specify any
    subset of congruence classes::

        sage: Primes(modulus=4, classes=[3])
        Set of prime numbers congruent to 3 modulo 4: 3, 7, 11, 19, ...
        sage: Primes(modulus=8, classes=[1, 3])
        Set of prime numbers congruent to 1, 3 modulo 8: 3, 11, 17, 19, ...

    If possible, the congruence conditions are simplified::

        sage: Primes(modulus=8, classes=[1, 5])
        Set of prime numbers congruent to 1 modulo 4: 5, 13, 17, 29, ...

    We can create a finite set of primes by passing in ``modulus=0``::

        sage: Primes(modulus=0, classes=[2, 3, 5, 11])
        Finite set of prime numbers: 2, 3, 5, 11

    We show various operations that can be performed on these sets::

        sage: P = Primes(modulus=4); P
        Set of prime numbers congruent to 1 modulo 4: 5, 13, 17, 29, ...
        sage: P.cardinality()
        +Infinity
        sage: P[:10]
        [5, 13, 17, 29, 37, 41, 53, 61, 73, 89]
        sage: P.next(500)
        509

    ::

        sage: Q = Primes(modulus=4, classes=[3])
        sage: PQ = P.union(Q)
        sage: PQ
        Set of all prime numbers with 2 excluded: 3, 5, 7, 11, ...
        sage: PQ.complement_in_primes()
        Finite set of prime numbers: 2
        sage: PQ.complement_in_primes().cardinality()
        1
    """
    @staticmethod
    def __classcall__(cls, modulus=1, classes=None, exceptions=None):
        """
        Normalize the input.

        INPUT:

        - ``modulus`` -- an integer (default: ``1``)

        - ``classes`` -- a list of integers (default: ``[1]``), the
          congruence classes (modulo ``modulus``) included in this
          set

        - ``exceptions`` -- a dictionary with items of the form
          ``x: b`` where ``x`` is an integer and ``b`` is a boolean;
          if ``b`` is ``True`` (resp. ``False``) then ``x`` is added
          to (resp. removed from) this set

        TESTS::

            sage: Primes(modulus=10)
            Set of prime numbers congruent to 1 modulo 5: 11, 31, 41, 61, ...
            sage: Primes(modulus=10) == Primes(modulus=5)
            True

            sage: Primes(modulus=9, classes=[1, 3, 4, 7])
            Set of prime numbers congruent to 1 modulo 3 with 3 included: 3, 7, 13, 19, ...

            sage: Primes(modulus=5, exceptions={7: True, 11: False})
            Set of prime numbers congruent to 1 modulo 5 with 7 included and 11 excluded: 7, 31, 41, 61, ...
        """
        modulus = ZZ(modulus)
        if modulus < 0:
            modulus = -modulus
        if classes is None:
            classes = [ZZ(1)]
        if exceptions is None:
            exceptions = {}
        if not isinstance(exceptions, dict):
            exceptions = dict(exceptions)

        if modulus == 0:
            for c in classes:
                exceptions[ZZ(c)] = True
            modulus = ZZ(1)
            classes = []

        # We replace each congruence of the form
        #   p = a (mod n) with gcd(a, n) > 1
        # (which includes at most one prime number)
        # with an exception
        indic = modulus * [False]
        for c in classes:
            indic[ZZ(c) % modulus] = True
        for c in range(modulus):
            if modulus.gcd(c) > 1:
                if indic[c]:
                    if c == 0:
                        if modulus not in exceptions:
                            exceptions[modulus] = True
                    elif c not in exceptions:
                        exceptions[ZZ(c)] = True
                indic[c] = None

        # We normalize the congruence conditions
        # by minimizing the modulus
        for p, mult in modulus.factor():
            while mult > 0:
                m = modulus // p
                add_true = []
                add_false = []
                add_excluded = []
                for c in range(m):
                    cs = [indic[c + m*i] for i in range(p) if indic[c + m*i] is not None]
                    if not cs:
                        pass
                    elif all(cs):
                        if m.gcd(c) == 1:
                            add_true.append(c)
                        for i in range(p):
                            j = c + m*i
                            if indic[j] is None:
                                if j == 0:
                                    add_excluded.append(modulus)
                                else:
                                    add_excluded.append(j)
                    elif not any(cs):
                        if m.gcd(c) == 1:
                            add_false.append(c)
                    else:
                        mult = 0
                        break
                else:
                    for c in add_true:
                        indic[c] = True
                    for c in add_false:
                        indic[c] = False
                    for x in add_excluded:
                        if x not in exceptions:
                            exceptions[x] = False
                    modulus = m
                    mult -= 1

        # We format the final result and make it hashable
        classes = tuple([c for c in range(modulus) if indic[c] is True])
        exceptions = [(ZZ(x), b) for x, b in exceptions.items()
                      if ZZ(x).is_prime() and (b != (indic[x % modulus] is True))]
        exceptions.sort()
        exceptions = tuple(exceptions)

        return super().__classcall__(cls, modulus, classes, exceptions)

    def __init__(self, modulus, classes, exceptions):
        r"""
        Initialize this set.

        TESTS::

            sage: P = Primes(modulus=4)
            sage: P.category()
            Category of facade infinite enumerated sets
            sage: TestSuite(P).run()

        ::

            sage: Q = Primes(modulus=0, classes=[2, 3, 5])
            sage: Q.category()
            Category of facade finite enumerated sets
            sage: TestSuite(Q).run()
        """
        if classes:
            category = InfiniteEnumeratedSets()
        else:
            category = FiniteEnumeratedSets()
        super().__init__(facade=ZZ, category=category)
        self._modulus = modulus
        self._classes = set(classes)
        self._exceptions = dict(exceptions)
        if classes:
            self._elements = []
        else:
            self._elements = [x for x, _ in exceptions]
            self._elements.sort()

    def congruence_classes(self):
        r"""
        Return the congruence classes selected in the subset
        of prime numbers.

        OUTPUT:

        A pair ``(modulus, list of classes)``

        EXAMPLES::

            sage: P = Primes(modulus=4)
            sage: P
            Set of prime numbers congruent to 1 modulo 4: 5, 13, 17, 29, ...
            sage: P.congruence_classes()
            (4, [1])

        If possible, the congruence classes are simplified::

            sage: P = Primes(modulus=10, classes=[1, 3])
            sage: P
            Set of prime numbers congruent to 1, 3 modulo 5: 3, 11, 13, 23, ...
            sage: P.congruence_classes()
            (5, [1, 3])

        If this subset is finite, the output of this method is always `(1, [])`.
        The elements of the subset can be retrieved using the method :meth:`list`
        or :meth:`included`::

            sage: P = Primes(modulus=0, classes=range(50))
            sage: P
            Finite set of prime numbers: 2, 3, 5, 7, ..., 43, 47
            sage: P.congruence_classes()
            (1, [])
            sage: list(P)
            [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47]

        .. SEEALSO::

            :meth:`included`, :meth:`excluded`
        """
        classes = list(self._classes)
        classes.sort()
        return (self._modulus, classes)

    def included(self):
        r"""
        Return the list of elements which are additionally included
        (that are, outside the congruence classes) to this set.

        EXAMPLES::

            sage: P = Primes(modulus=4)
            sage: P
            Set of prime numbers congruent to 1 modulo 4: 5, 13, 17, 29, ...
            sage: P.included()
            []

        ::

            sage: Q = P.include(2)
            sage: Q
            Set of prime numbers congruent to 1 modulo 4 with 2 included: 2, 5, 13, 17, ...
            sage: Q.included()
            [2]

        .. SEEALSO::

            :meth:`excluded`, :meth:`congruence_classes`
        """
        included = [x for x, b in self._exceptions.items() if b]
        included.sort()
        return included

    def excluded(self):
        r"""
        Return the list of elements which are excluded, that are the
        elements in the congruence classes defining this subset but
        not in this subset.

        EXAMPLES::

            sage: P = Primes(modulus=4)
            sage: P
            Set of prime numbers congruent to 1 modulo 4: 5, 13, 17, 29, ...
            sage: P.excluded()
            []

        ::

            sage: Q = P.exclude(5)
            sage: Q
            Set of prime numbers congruent to 1 modulo 4 with 5 excluded: 13, 17, 29, 37, ...
            sage: Q.excluded()
            [5]

        .. SEEALSO::

            :meth:`included`, :meth:`congruence_classes`
        """
        excluded = [x for x, b in self._exceptions.items() if not b]
        excluded.sort()
        return excluded

    def _repr_(self):
        r"""
        Return a string representation of this subset.

        TESTS::

            sage: Primes(modulus=4).include(2).exclude(5)  # indirect doctest
            Set of prime numbers congruent to 1 modulo 4 with 2 included and 5 excluded: 2, 13, 17, 29, ...

            sage: E = Primes(modulus=0)
            sage: E  # indirect doctest
            Empty set of prime numbers

            sage: E.include(range(50), check=False)  # indirect doctest
            Finite set of prime numbers: 2, 3, 5, 7, ..., 43, 47
        """
        _, classes = self.congruence_classes()
        included = self.included()
        excluded = self.excluded()
        if not classes:
            if not included:
                return "Empty set of prime numbers"
            return "Finite set of prime numbers: %s" % _repr_items(included)
        if self._modulus == 1:
            s = "Set of all prime numbers"
        else:
            s = "Set of prime numbers congruent to %s modulo %s" % (_repr_items(classes), self._modulus)
        if included:
            s += " with %s included" % _repr_items(included)
        if excluded:
            if not included:
                s += " with %s excluded" % _repr_items(excluded)
            else:
                s += " and %s excluded" % _repr_items(excluded)
        s += ": %s, ..." % (", ".join([str(n) for n in self[:4]]))
        return s

    def __contains__(self, x):
        r"""
        Return ``True`` if `x` is in this set; ``False`` otherwise.

        INPUT:

        - ``x`` -- an integer

        EXAMPLES::

            sage: P = Primes(modulus=4)
            sage: 3 in P
            False
            sage: 9 in P
            False
            sage: 13 in P
            True

        TESTS::

            sage: x in P
            False
        """
        try:
            if x not in ZZ:
                return False
        except TypeError:
            return False
        x = ZZ(x)
        if not x.is_prime():
            return False
        e = self._exceptions.get(x, None)
        return (e is True) or (e is None and x % self._modulus in self._classes)

    def cardinality(self):
        r"""
        Return the cardinality of this set.

        EXAMPLES::

            sage: P = Primes(modulus=4); P
            Set of prime numbers congruent to 1 modulo 4: 5, 13, 17, 29, ...
            sage: P.cardinality()
            +Infinity

        ::

            sage: P = Primes(modulus=4, classes=[2]); P
            Finite set of prime numbers: 2
            sage: P.cardinality()
            1
        """
        if self.is_finite():
            return ZZ(len(self._elements))
        return infinity

    def first(self, n=None):
        r"""
        Return the first element in this set.

        EXAMPLES::

            sage: P = Primes(modulus=5); P
            Set of prime numbers congruent to 1 modulo 5: 11, 31, 41, 61, ...
            sage: P.first()
            11
        """
        if self.is_empty():
            raise EmptySetError
        return self.next(1)

    def next(self, x):
        r"""
        Return the smallest element in this set strictly
        greater than ``x``.

        INPUT:

        - ``x`` -- an integer

        EXAMPLES::

            sage: P = Primes(modulus=5); P
            Set of prime numbers congruent to 1 modulo 5: 11, 31, 41, 61, ...
            sage: P.next(1000)
            1021

        If there is no element greater than the given bound, an
        error is raised::

            sage: P = Primes(modulus=0, classes=[2, 5]); P
            Finite set of prime numbers: 2, 5
            sage: P.next(10)
            Traceback (most recent call last):
            ...
            ValueError: no element greater that 10 in this set
        """
        x = ZZ(x)
        if self._classes:
            while True:
                x = x.next_prime()
                e = self._exceptions.get(x, None)
                if (e is True) or (e is None and x % self._modulus in self._classes):
                    return x

        if not self._elements or x >= self._elements[-1]:
            raise ValueError("no element greater that %s in this set" % x)
        min = 0
        max = len(self._elements)
        while min < max:
            i = (min + max) // 2
            if self._elements[i] <= x:
                min = i + 1
            if self._elements[i] > x:
                max = i
        return self._elements[min]

    def _an_element_(self):
        r"""
        Return an element in this set.

        EXAMPLES::

            sage: P = Primes()
            sage: P.an_element()  # indirect doctest
            43

        If the set is empty, an error is raised::

            sage: P = Primes(modulus=12, classes=[4])
            sage: P.an_element()  # indirect doctest
            Traceback (most recent call last):
            ...
            ValueError: this set is empty
        """
        if self.is_finite():
            if self._elements:
                return self._elements[0]
            raise ValueError("this set is empty")
        return self.next(42)

    def unrank(self, n):
        r"""
        Return the ``n``-th element of this set.

        INPUT:

        - ``n`` -- an integer

        EXAMPLES::

            sage: P = Primes(modulus=5); P
            Set of prime numbers congruent to 1 modulo 5: 11, 31, 41, 61, ...
            sage: P[0]   # indirect doctest
            11
            sage: P[10]  # indirect doctest
            211

        If there is less than `n` elements in this set, an error
        is raised::

            sage: P = Primes(modulus=0, classes=[2, 5]); P
            Finite set of prime numbers: 2, 5
            sage: P[1]
            5
            sage: P[2]  # indirect doctest
            Traceback (most recent call last):
            ...
            IndexError: this set has not enough elements

        TESTS::

            sage: P.unrank(-1)
            Traceback (most recent call last):
            ...
            IndexError: index must be nonnegative
        """
        if n < 0:
            raise IndexError("index must be nonnegative")
        if self.is_finite():
            if len(self._elements) <= n:
                raise IndexError("this set has not enough elements")
            return self._elements[n]
        if self._elements:
            x = self._elements[-1]
        else:
            x = 1
        while len(self._elements) <= n:
            x = self.next(x)
            self._elements.append(x)
        return self._elements[n]

    def is_empty(self):
        r"""
        Return ``True`` if this set is empty; ``False`` otherwise.

        EXAMPLES::

            sage: P = Primes(modulus=6); P
            Set of prime numbers congruent to 1 modulo 3: 7, 13, 19, 31, ...
            sage: P.is_empty()
            False

        ::

            sage: P = Primes(modulus=6, classes=[6]); P
            Empty set of prime numbers
            sage: P.is_empty()
            True

        .. SEEALSO::

            :meth:`is_finite`, :meth:`is_cofinite`
        """
        return not bool(self._classes) and not bool(self._exceptions)

    def is_finite(self):
        r"""
        Return ``True`` if this set is finite; ``False`` otherwise.

        EXAMPLES::

            sage: P = Primes(modulus=5); P
            Set of prime numbers congruent to 1 modulo 5: 11, 31, 41, 61, ...
            sage: P.is_finite()
            False

        ::

            sage: P = Primes(modulus=0, classes=[2, 5]); P
            Finite set of prime numbers: 2, 5
            sage: P.is_finite()
            True

        .. SEEALSO::

            :meth:`is_empty`, :meth:`is_cofinite`
        """
        return not bool(self._classes)

    def is_cofinite(self):
        r"""
        Return ``True`` if this set is cofinite in the set
        of all prime numbers; ``False`` otherwise.

        EXAMPLES::

            sage: P = Primes(modulus=4); P
            Set of prime numbers congruent to 1 modulo 4: 5, 13, 17, 29, ...
            sage: P.is_cofinite()
            False

        ::

            sage: P = Primes(modulus=4, classes=[1, 3]); P
            Set of all prime numbers with 2 excluded: 3, 5, 7, 11, ...
            sage: P.is_cofinite()
            True

        .. SEEALSO::

            :meth:`is_empty`, :meth:`is_finite`
        """
        return self._modulus == 1 and bool(self._classes)

    def density(self):
        r"""
        Return the density of this set in the set of all
        prime numbers.

        EXAMPLES::

            sage: P = Primes(modulus=4); P
            Set of prime numbers congruent to 1 modulo 4: 5, 13, 17, 29, ...
            sage: P.density()
            1/2
        """
        return len(self._classes) / euler_phi(self._modulus)

    def include(self, elements, check=True):
        r"""
        Return this set with the integers in ``elements`` included.

        INPUT:

        - ``elements`` -- an integer, or a tuple/list of integers

        - ``check`` -- a boolean (default: ``True``); if ``False``,
          do not raise an error if we try to add composite numbers

        EXAMPLES::

            sage: P = Primes(modulus=5); P
            Set of prime numbers congruent to 1 modulo 5: 11, 31, 41, 61, ...
            sage: P.include(2)
            Set of prime numbers congruent to 1 modulo 5 with 2 included: 2, 11, 31, 41, ...
            sage: P.include([2, 3])
            Set of prime numbers congruent to 1 modulo 5 with 2, 3 included: 2, 3, 11, 31, ...

        If we try to include an element which is already in the set,
        nothing changes::

            sage: P.include(11)
            Set of prime numbers congruent to 1 modulo 5: 11, 31, 41, 61, ...

        Trying to include a composite number results in an error::

            sage: P.include(10)
            Traceback (most recent call last):
            ...
            ValueError: 10 is not a prime number

        We can avoid this by passing in ``check=False``; in this case,
        composite numbers are however not added to the set.
        This behavior can be convenient if one wants to add all prime
        numbers in a range::

            sage: P.include(range(20, 30), check=False)
            Set of prime numbers congruent to 1 modulo 5 with 23, 29 included: 11, 23, 29, 31, ...

        .. SEEALSO::

            :meth:`exclude`
        """
        if elements in ZZ:
            elements = [elements]
        exceptions = self._exceptions.copy()
        for x in elements:
            x = ZZ(x)
            if check and not x.is_prime():
                raise ValueError("%s is not a prime number" % x)
            exceptions[x] = True
        return Primes(self._modulus, self._classes, exceptions)

    def exclude(self, elements):
        r"""
        Return this set with the integers in ``elements`` excluded.

        INPUT:

        - ``elements`` -- an integer, or a tuple/list of integers

        EXAMPLES::

            sage: P = Primes(modulus=5); P
            Set of prime numbers congruent to 1 modulo 5: 11, 31, 41, 61, ...
            sage: P.exclude(11)
            Set of prime numbers congruent to 1 modulo 5 with 11 excluded: 31, 41, 61, 71, ...
            sage: P.exclude([11, 31])
            Set of prime numbers congruent to 1 modulo 5 with 11, 31 excluded: 41, 61, 71, 101, ...

        If we try to exclude an element which is not in the set,
        nothing changes::

            sage: P.exclude(2)
            Set of prime numbers congruent to 1 modulo 5: 11, 31, 41, 61, ...

        .. SEEALSO::

            :meth:`include`
        """
        if elements in ZZ:
            elements = [elements]
        exceptions = self._exceptions.copy()
        for x in elements:
            exceptions[x] = False
        return Primes(self._modulus, self._classes, exceptions)

    def complement_in_primes(self):
        r"""
        Return the complement of this set in the set of all prime
        numbers.

        EXAMPLES::

            sage: P = Primes(modulus=4); P
            Set of prime numbers congruent to 1 modulo 4: 5, 13, 17, 29, ...
            sage: Q = P.complement_in_primes(); Q
            Set of prime numbers congruent to 3 modulo 4 with 2 included: 2, 3, 7, 11, ...

        We check that the union of `P` and `Q` is the whole set of
        prime numbers::

            sage: P.union(Q)
            Set of all prime numbers: 2, 3, 5, 7, ...

        and that the intersection is empty::

            sage: P.intersection(Q)
            Empty set of prime numbers

        .. SEEALSO::

            :meth:`intersection`, :meth:`union`
        """
        modulus = self._modulus
        classes = [c for c in range(modulus)
                   if c % self._modulus not in self._classes]
        exceptions = {x: not b for x, b in self._exceptions.items()}
        return Primes(modulus, classes, exceptions)

    def intersection(self, other):
        r"""
        Return the intersection of this set with ``other``.

        INPUT:

        - ``other`` -- a subset of the set of prime numbers

        EXAMPLES::

            sage: P = Primes(modulus=5); P
            Set of prime numbers congruent to 1 modulo 5: 11, 31, 41, 61, ...
            sage: Q = Primes(modulus=3, classes=[2]); Q
            Set of prime numbers congruent to 2 modulo 3: 2, 5, 11, 17, ...
            sage: P.intersection(Q)
            Set of prime numbers congruent to 11 modulo 15: 11, 41, 71, 101, ...

        It is also possible to take the intersection with a range::

            sage: P.intersection(range(100))
            Finite set of prime numbers: 11, 31, 41, 61, 71

        TESTS::

            sage: P = Primes(modulus=5, exceptions={5: True, 11: False}); P
            Set of prime numbers congruent to 1 modulo 5 with 5 included and 11 excluded: 5, 31, 41, 61, ...
            sage: Q = Primes(modulus=4, exceptions={13: False, 11: True}); Q
            Set of prime numbers congruent to 1 modulo 4 with 11 included and 13 excluded: 5, 11, 17, 29, ...
            sage: P.intersection(Q)
            Set of prime numbers congruent to 1 modulo 20 with 5 included: 5, 41, 61, 101, ...

        ::

            sage: P.intersection(NN) == P
            True
            sage: P.intersection(ZZ) == P
            True
            sage: P.intersection(RR)
            Traceback (most recent call last):
            ...
            NotImplementedError: intersection with general infinite sets is not implemented

            sage: P = Primes(modulus=0, classes=range(30))
            sage: P.intersection(reversed([13, 7, 11, 37]))
            Finite set of prime numbers: 7, 11, 13


        .. SEEALSO::

            :meth:`complement_in_primes`, :meth:`union`
        """
        if other is NN or other is ZZ:
            return self
        if isinstance(other, Primes):
            modulus = self._modulus.lcm(other._modulus)
            classes = [c for c in range(modulus)
                       if (c % self._modulus in self._classes
                       and c % other._modulus in other._classes)]
            exceptions = {x: b for x, b in self._exceptions.items()
                          if not b or x in other}
            exceptions.update((x, b) for x, b in other._exceptions.items()
                              if not b or x in self)
        else:
            modulus = 1
            classes = []
            if isinstance(other, range) and other.step == 1:
                exceptions = {}
                x = other.start - 1
                while True:
                    try:
                        x = self.next(x)
                    except ValueError:
                        break
                    if x >= other.stop:
                        break
                    exceptions[x] = True
            elif self.is_finite() and hasattr(other, "__contains__"):
                # this would not work reliably if ``other`` does not
                # implement ``__contains__``, because ``x in other``
                # then consumes ``other``
                exceptions = {x: True for x in self if x in other}
            else:
                if hasattr(other, "is_finite") and not other.is_finite():
                    raise NotImplementedError("intersection with general infinite sets is not implemented")
                # if other is infinite but does not know it, this will loop forever
                exceptions = {x: True for x in other if x in self}
        return Primes(modulus, classes, exceptions)

    def union(self, other):
        r"""
        Return the union of this set and ``other``.

        INPUT:

        - ``other`` -- a subset of the set of prime numbers

        EXAMPLES::

            sage: P = Primes(modulus=5); P
            Set of prime numbers congruent to 1 modulo 5: 11, 31, 41, 61, ...
            sage: Q = Primes(modulus=3, classes=[2]); Q
            Set of prime numbers congruent to 2 modulo 3: 2, 5, 11, 17, ...
            sage: P.union(Q)
            Set of prime numbers congruent to 1, 2, 8, 11, 14 modulo 15 with 5 included: 2, 5, 11, 17, ...

        TESTS::

            sage: P = Primes(modulus=5, exceptions={5: True, 11: False}); P
            Set of prime numbers congruent to 1 modulo 5 with 5 included and 11 excluded: 5, 31, 41, 61, ...
            sage: Q = Primes(modulus=4, exceptions={13: False, 11: True}); Q
            Set of prime numbers congruent to 1 modulo 4 with 11 included and 13 excluded: 5, 11, 17, 29, ...
            sage: P.union(Q)
            Set of prime numbers congruent to 1, 9, 11, 13, 17 modulo 20 with 5 included and 13 excluded: 5, 11, 17, 29, ...

        ::

            sage: P.union(NN) == NN
            True
            sage: P.union(ZZ) == ZZ
            True
            sage: P.union(RR)
            Traceback (most recent call last):
            ...
            NotImplementedError: union with general infinite sets is not implemented

        .. SEEALSO::

            :meth:`complement_in_primes`, :meth:`intersection`
        """
        if other is NN:
            return NN
        if other is ZZ:
            return ZZ
        if isinstance(other, Primes):
            modulus = self._modulus.lcm(other._modulus)
            classes = [c for c in range(modulus)
                       if (c % self._modulus in self._classes
                        or c % other._modulus in other._classes)]
            exceptions = {x: b for x, b in self._exceptions.items()
                          if b or x not in other}
            exceptions.update((x, b) for x, b in other._exceptions.items()
                              if b or x not in self)
        else:
            # we try to enumerate the elements of "other"
            if hasattr(other, "is_finite") and not other.is_finite():
                raise NotImplementedError("union with general infinite sets is not implemented")
            modulus = self._modulus
            classes = self._classes
            exceptions = self._exceptions.copy()
            for x in other:
                x = ZZ(x)
                if x.is_prime():
                    exceptions[x] = True
                else:
                    raise NotImplementedError("the result of the union is a subset of the set of prime numbers")
        return Primes(modulus, classes, exceptions)

    def is_almost_equal(self, other):
        r"""
        Return whether this set only differs from ``other``
        by a finite set.

        INPUT:

        - ``other`` -- a subset of the set of prime numbers

        EXAMPLES::

            sage: P = Primes(modulus=20, classes=[1, 2]); P
            Set of prime numbers congruent to 1 modulo 20 with 2 included: 2, 41, 61, 101, ...
            sage: Q = Primes(modulus=20, classes=[1, 5]); Q
            Set of prime numbers congruent to 1 modulo 20 with 5 included: 5, 41, 61, 101, ...
            sage: P.is_almost_equal(Q)
            True

        ::

            sage: R = Primes(modulus=10); R
            Set of prime numbers congruent to 1 modulo 5: 11, 31, 41, 61, ...
            sage: P.is_almost_equal(R)
            False
        """
        if not isinstance(other, Primes):
            return False  # or raise an error?
        return self._modulus == other._modulus and self._classes == other._classes

    def is_subset(self, other, almost=False):
        r"""
        Return whether this set is a subset of ``other``.

        INPUT:

        - ``other`` -- a subset of the set of prime numbers

        - ``almost`` -- a boolean (default: ``False``); if ``True``,
          the inclusion is only checked up to a finite set

        EXAMPLES::

            sage: P = Primes(modulus=4); P
            Set of prime numbers congruent to 1 modulo 4: 5, 13, 17, 29, ...
            sage: Q = Primes(modulus=8); Q
            Set of prime numbers congruent to 1 modulo 8: 17, 41, 73, 89, ...
            sage: P.is_subset(Q)
            False
            sage: Q.is_subset(P)
            True

        When ``almost=True``, the inclusion is only checked up to a
        finite set::

            sage: Q2 = Q.include(2); Q2
            Set of prime numbers congruent to 1 modulo 8 with 2 included: 2, 17, 41, 73, ...
            sage: Q2.is_subset(P)
            False
            sage: Q2.is_subset(P, almost=True)
            True

        TESTS::

            sage: P.is_subset(ZZ)
            True

        .. SEEALSO::

            :meth:`is_superset`, :meth:`is_disjoint`, :meth:`is_almost_equal`
        """
        if other is ZZ:
            return True
        P = self.intersection(other)
        if almost:
            return P.is_almost_equal(self)
        return P == self

    def is_superset(self, other, almost=False):
        r"""
        Return whether this set contains the set ``other`` as a subset.

        INPUT:

        - ``other`` -- a subset of the set of prime numbers

        - ``almost`` -- a boolean (default: ``False``); if ``True``,
          the inclusion is only checked up to a finite set

        EXAMPLES::

            sage: P = Primes(modulus=4); P
            Set of prime numbers congruent to 1 modulo 4: 5, 13, 17, 29, ...
            sage: Q = Primes(modulus=8); Q
            Set of prime numbers congruent to 1 modulo 8: 17, 41, 73, 89, ...
            sage: P.is_superset(Q)
            True
            sage: Q.is_superset(P)
            False

        When ``almost=True``, the inclusion is only checked up to a
        finite set::

            sage: Q2 = Q.include(2); Q2
            Set of prime numbers congruent to 1 modulo 8 with 2 included: 2, 17, 41, 73, ...
            sage: P.is_superset(Q2)
            False
            sage: P.is_superset(Q2, almost=True)
            True

        TESTS::

            sage: P.is_superset(ZZ)
            False

        .. SEEALSO::

            :meth:`is_subset`, :meth:`is_disjoint`, :meth:`is_almost_equal`
        """
        if other is ZZ:
            return False
        P = self.intersection(other)
        if almost:
            return P.is_almost_equal(other)
        return P == other

    def is_disjoint(self, other, almost=False):
        r"""
        Return whether the intersection of this set with ``other``
        is empty (or finite, if ``almost`` is ``True``).

        INPUT:

        - ``other`` -- a subset of the set of prime numbers

        - ``almost`` -- a boolean (default: ``False``)

        EXAMPLES::

            sage: P = Primes(modulus=4); P
            Set of prime numbers congruent to 1 modulo 4: 5, 13, 17, 29, ...
            sage: Q = Primes(modulus=4, classes=[3]); Q
            Set of prime numbers congruent to 3 modulo 4: 3, 7, 11, 19, ...
            sage: P.is_disjoint(Q)
            True

        ::

            sage: R = Primes(modulus=5, classes=[3]); R
            Set of prime numbers congruent to 3 modulo 5: 3, 13, 23, 43, ...
            sage: P.is_disjoint(R)
            False
            sage: Q.is_disjoint(R)
            False

        We illustrate the behavior when ``almost=True``::

            sage: Q5 = Q.include(5); Q5
            Set of prime numbers congruent to 3 modulo 4 with 5 included: 3, 5, 7, 11, ...
            sage: P.is_disjoint(Q5)
            False
            sage: P.is_disjoint(Q5, almost=True)
            True

        .. SEEALSO::

            :meth:`is_subset`, :meth:`is_disjoint`, :meth:`is_almost_equal`
        """
        P = self.intersection(other)
        if almost:
            return P.is_finite()
        return P.is_empty()



"""
`p`-adic Fixed-Mod Element

Elements of `p`-adic Rings with Fixed Modulus

AUTHORS:

- David Roe
- Genya Zaytman: documentation
- David Harvey: doctests
"""

#*****************************************************************************
#       Copyright (C) 2007-2013 David Roe <roed.math@gmail.com>
#                               William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

include "sage/libs/linkages/padics/mpz.pxi"
include "FM_template.pxi"

from sage.libs.pari.convert_gmp cimport new_gen_from_padic
from sage.rings.finite_rings.integer_mod import Mod

cdef extern from "transcendental.c":
    cdef void padiclog(mpz_t ans, const mpz_t a, unsigned long p, unsigned long prec, const mpz_t modulo)
    cdef void padicexp(mpz_t ans, const mpz_t a, unsigned long p, unsigned long prec, const mpz_t modulo)
    cdef void padicexp_Newton(mpz_t ans, const mpz_t a, unsigned long p, unsigned long prec, unsigned long precinit, const mpz_t modulo)

cdef class PowComputer_(PowComputer_base):
    """
    A PowComputer for a fixed-modulus `p`-adic ring.
    """
    def __init__(self, Integer prime, long cache_limit, long prec_cap, long ram_prec_cap, bint in_field):
        """
        Initialization.

        EXAMPLES::

            sage: R = ZpFM(5)
            sage: type(R.prime_pow)
            <class 'sage.rings.padics.padic_fixed_mod_element.PowComputer_'>
            sage: R.prime_pow._prec_type
            'fixed-mod'
        """
        self._prec_type = 'fixed-mod'
        PowComputer_base.__init__(self, prime, cache_limit, prec_cap, ram_prec_cap, in_field)

cdef class pAdicFixedModElement(FMElement):
    r"""
    INPUT:

    - ``parent`` -- a ``pAdicRingFixedMod`` object

    - ``x`` -- input data to be converted into the parent

    - ``absprec`` -- ignored; for compatibility with other `p`-adic rings

    - ``relprec`` -- ignored; for compatibility with other `p`-adic rings

    .. NOTE::

        The following types are currently supported for x:

        - Integers
        - Rationals -- denominator must be relatively prime to `p`
        - FixedMod `p`-adics
        - Elements of ``IntegerModRing(p^k)`` for ``k`` less than or equal
          to the modulus

        The following types should be supported eventually:

        - Finite precision `p`-adics
        - Lazy `p`-adics
        - Elements of local extensions of THIS `p`-adic ring that actually
          lie in `\ZZ_p`

    EXAMPLES::

        sage: R = Zp(5, 20, 'fixed-mod', 'terse')

    Construct from integers::

        sage: R(3)
        3
        sage: R(75)
        75
        sage: R(0)
        0

        sage: R(-1)
        95367431640624
        sage: R(-5)
        95367431640620

    Construct from rationals::

        sage: R(1/2)
        47683715820313
        sage: R(-7875/874)
        9493096742250
        sage: R(15/425)
        Traceback (most recent call last):
        ...
        ValueError: p divides denominator

    Construct from IntegerMod::

        sage: R(Integers(125)(3))
        3
        sage: R(Integers(5)(3))
        3
        sage: R(Integers(5^30)(3))
        3
        sage: R(Integers(5^30)(1+5^23))
        1
        sage: R(Integers(49)(3))
        Traceback (most recent call last):
        ...
        TypeError: p does not divide modulus 49

        sage: R(Integers(48)(3))
        Traceback (most recent call last):
        ...
        TypeError: p does not divide modulus 48

    Some other conversions::

        sage: R(R(5))
        5

    .. TODO:: doctests for converting from other types of `p`-adic rings
    """
    def lift(self):
        r"""
        Return an integer congruent to ``self`` modulo the precision.

        .. WARNING::

            Since fixed modulus elements don't track their precision,
            the result may not be correct modulo
            `i^{\mathrm{prec_cap}}` if the element was defined by
            constructions that lost precision.

        EXAMPLES::

            sage: R = Zp(7,4,'fixed-mod'); a = R(8); a.lift()
            8
            sage: type(a.lift())
            <class 'sage.rings.integer.Integer'>
        """
        return self.lift_c()

    cdef lift_c(self):
        r"""
        Return an integer congruent to this element modulo the precision.

        .. WARNING::

            Since fixed modulus elements don't track their precision,
            the result may not be correct modulo
            `i^{\mbox{prec_cap}}` if the element was defined by
            constructions that lost precision.

        EXAMPLES::

            sage: R = ZpFM(7,4); a = R(8); a.lift()  # indirect doctest
            8
        """
        cdef Integer ans = PY_NEW(Integer)
        mpz_set(ans.value, self.value)
        return ans

    def __pari__(self):
        """
        Conversion to PARI.

        EXAMPLES::

            sage: R = ZpCA(5)
            sage: pari(R(1777))  # indirect doctest                                     # needs sage.libs.pari
            2 + 5^2 + 4*5^3 + 2*5^4 + O(5^20)
        """
        return self._to_gen()

    cdef pari_gen _to_gen(self):
        """
        Convert ``self`` to an equivalent pari element.

        EXAMPLES::

            sage: R = ZpFM(5, 10); a = R(17); pari(a)  # indirect doctest
            2 + 3*5 + O(5^10)
            sage: pari(R(0))                                                            # needs sage.libs.pari
            O(5^10)
            sage: pari(R(0,5))                                                          # needs sage.libs.pari
            O(5^10)
            sage: pari(R(0)).debug()                                                    # needs sage.libs.pari
            [&=...] PADIC(lg=5):... (precp=0,valp=10):... ... ... ...
                p : [&=...] INT(lg=3):... (+,lgefint=3):... ...
              p^l : [&=...] INT(lg=3):... (+,lgefint=3):... ...
                I : gen_0

        This checks that :issue:`15653` is fixed::

            sage: x = polygen(ZpFM(3,10))
            sage: (x^3 + x + 1).__pari__().poldisc()
            2 + 3 + 2*3^2 + 3^3 + 2*3^4 + 2*3^5 + 2*3^6 + 2*3^7 + 2*3^8 + 2*3^9 + O(3^10)
        """
        cdef long val
        # Let val be the valuation of self, holder (defined in the
        # linkage file) be the unit part.
        if mpz_sgn(self.value) == 0:
            # Special case for zero: maximal valuation and 0 unit part
            val = self.prime_pow.prec_cap
            mpz_set_ui(holder.value, 0)
        else:
            val = mpz_remove(holder.value, self.value, self.prime_pow.prime.value)
        return new_gen_from_padic(val, self.prime_pow.prec_cap - val,
                                  self.prime_pow.prime.value,
                                  self.prime_pow.pow_mpz_t_tmp(self.prime_pow.prec_cap - val),
                                  holder.value)

    def _integer_(self, Z=None):
        r"""
        Return an integer congruent to ``self`` modulo the precision.

        .. WARNING::

            Since fixed modulus elements don't track their precision,
            the result may not be correct modulo
            `p^{\mathrm{prec_cap}}` if the element was defined by
            constructions that lost precision.

        EXAMPLES::

            sage: R = ZpFM(5); R(-1)._integer_()
            95367431640624
        """
        return self.lift_c()

    def residue(self, absprec=1, field=None, check_prec=False):
        r"""
        Reduce ``self`` modulo `p^\mathrm{absprec}`.

        INPUT:

        - ``absprec`` -- integer (default: 1)

        - ``field`` -- boolean (default: ``None``); whether to return an
          element of GF(p) or Zmod(p)

        - ``check_prec`` -- boolean (default: ``False``); no effect (for
          compatibility with other types)

        OUTPUT:

        This element reduced modulo `p^\mathrm{absprec}` as an element of
        `\ZZ/p^\mathrm{absprec}\ZZ`.

        EXAMPLES::

            sage: R = Zp(7,4,'fixed-mod')
            sage: a = R(8)
            sage: a.residue(1)
            1

        This is different from applying ``% p^n`` which returns an element in
        the same ring::

            sage: b = a.residue(2); b
            8
            sage: b.parent()
            Ring of integers modulo 49
            sage: c = a % 7^2; c
            1 + 7
            sage: c.parent()
            7-adic Ring of fixed modulus 7^4

        TESTS::

            sage: R = Zp(7,4,'fixed-mod')
            sage: a = R(8)
            sage: a.residue(0)
            0
            sage: a.residue(-1)
            Traceback (most recent call last):
            ...
            ValueError: cannot reduce modulo a negative power of p
            sage: a.residue(5)
            8

            sage: a.residue(field=True).parent()
            Finite Field of size 7

        .. SEEALSO::

            :meth:`_mod_`
        """
        cdef Integer selfvalue, modulus
        cdef long aprec
        if not isinstance(absprec, Integer):
            absprec = Integer(absprec)
        if absprec < 0:
            raise ValueError("cannot reduce modulo a negative power of p")
        if field is None:
            field = (absprec == 1)
        elif field and absprec != 1:
            raise ValueError("field keyword may only be set at precision 1")
        if mpz_fits_slong_p((<Integer>absprec).value) == 0:
            raise ValueError("absolute precision does not fit in a long")
        aprec = mpz_get_si((<Integer>absprec).value)
        modulus = PY_NEW(Integer)
        mpz_set(modulus.value, self.prime_pow.pow_mpz_t_tmp(aprec))
        selfvalue = PY_NEW(Integer)
        mpz_set(selfvalue.value, self.value)
        if field:
            from sage.rings.finite_rings.finite_field_constructor import GF
            return GF(self.parent().prime())(selfvalue)
        else:
            return Mod(selfvalue, modulus)

    def multiplicative_order(self):
        r"""
        Return the minimum possible multiplicative order of ``self``.

        OUTPUT:

        an integer -- the multiplicative order of this element.  This is the
        minimum multiplicative order of all elements of `\ZZ_p` lifting this
        element to infinite precision.

         EXAMPLES::

            sage: R = ZpFM(7, 6)
            sage: R(1/3)
            5 + 4*7 + 4*7^2 + 4*7^3 + 4*7^4 + 4*7^5
            sage: R(1/3).multiplicative_order()
            +Infinity
            sage: R(7).multiplicative_order()
            +Infinity
            sage: R(1).multiplicative_order()
            1
            sage: R(-1).multiplicative_order()
            2
            sage: R.teichmuller(3).multiplicative_order()
            6
        """
        cdef mpz_t tmp
        cdef Integer ans
        if mpz_divisible_p(self.value, self.prime_pow.prime.value):
            return infinity
        if mpz_cmp_ui(self.value, 1) == 0:
            ans = PY_NEW(Integer)
            mpz_set_ui(ans.value, 1)
            return ans
        mpz_init(tmp)
        mpz_sub_ui(tmp, self.prime_pow.pow_mpz_t_top(), 1)
        if mpz_cmp(self.value, tmp) == 0:
            ans = PY_NEW(Integer)
            mpz_set_ui(ans.value, 2)
            return ans
        # check if self is an approximation to a teichmuller lift:
        mpz_powm(tmp, self.value, self.prime_pow.prime.value, self.prime_pow.pow_mpz_t_top())
        if mpz_cmp(tmp, self.value) == 0:
            mpz_clear(tmp)
            return self.residue(1).multiplicative_order()
        else:
            mpz_clear(tmp)
            return infinity

    def _log_binary_splitting(self, aprec, mina=0):
        r"""
        Return ``\log(self)`` for ``self`` equal to 1 in the residue field.

        This is a helper method for :meth:`log`.
        It uses a fast binary splitting algorithm.

        INPUT:

        - ``aprec`` -- integer; the precision to which the result is
          correct. ``aprec`` must not exceed the precision cap of the ring over
          which this element is defined.
        - ``mina`` -- integer (default: 0); the series will check `n` up to
          this valuation (and beyond) to see if they can contribute to the
          series

        .. NOTE::

            The function does not check that its argument ``self`` is
            1 in the residue field. If this assumption is not fulfilled
            the behaviour of the function is not specified.

        ALGORITHM:

        1. Raise `u` to the power `p^v` for a suitable `v` in order
           to make it closer to 1. (`v` is chosen such that `p^v` is
           close to the precision.)

        2. Write

        .. MATH::

            u^{p-1} = \prod_{i=1}^\infty (1 - a_i p^{(v+1)*2^i})

        with `0 \leq a_i < p^{(v+1)*2^i}` and compute
        `\log(1 - a_i p^{(v+1)*2^i})` using the standard Taylor expansion

        .. MATH::

            \log(1 - x) = -x - 1/2 x^2 - 1/3 x^3 - 1/4 x^4 - 1/5 x^5 - \cdots

        together with a binary splitting method.

        3. Divide the result by `p^v`

        The complexity of this algorithm is quasi-linear.

        EXAMPLES::

            sage: r = Qp(5,prec=4)(6)
            sage: r._log_binary_splitting(2)
            5 + O(5^2)
            sage: r._log_binary_splitting(4)
            5 + 2*5^2 + 4*5^3 + O(5^4)
            sage: r._log_binary_splitting(100)
            5 + 2*5^2 + 4*5^3 + O(5^4)

            sage: r = Zp(5,prec=4,type='fixed-mod')(6)
            sage: r._log_binary_splitting(5)
            5 + 2*5^2 + 4*5^3
        """
        cdef unsigned long p
        cdef unsigned long prec = min(aprec, self.prime_pow.prec_cap)
        cdef pAdicFixedModElement ans

        if mpz_fits_slong_p(self.prime_pow.prime.value) == 0:
            raise NotImplementedError("the prime %s does not fit in a long" % self.prime_pow.prime)
        p = self.prime_pow.prime

        ans = self._new_c()
        sig_on()
        padiclog(ans.value, self.value, p, prec, self.prime_pow.pow_mpz_t_tmp(prec))
        sig_off()
        return ans

    def _exp_binary_splitting(self, aprec):
        r"""
        Compute the exponential power series of this element.

        This is a helper method for :meth:`exp`.

        INPUT:

        - ``aprec`` -- integer; the precision to which to compute the
          exponential

        .. NOTE::

            The function does not check that its argument ``self`` is
            the disk of convergence of ``exp``. If this assumption is not
            fulfilled the behaviour of the function is not specified.

        ALGORITHM:

        Write

        .. MATH::

            self = \sum_{i=1}^\infty a_i p^{2^i}

        with `0 \leq a_i < p^{2^i}` and compute
        `\exp(a_i p^{2^i})` using the standard Taylor expansion

        .. MATH::

            \exp(x) = 1 + x + x^2/2 + x^3/6 + x^4/24 + \cdots

        together with a binary splitting method.

        The binary complexity of this algorithm is quasi-linear.

        EXAMPLES::

            sage: R = Zp(7,5)
            sage: x = R(7)
            sage: x.exp(algorithm='binary_splitting')   # indirect doctest
            1 + 7 + 4*7^2 + 2*7^3 + O(7^5)
        """
        cdef unsigned long p
        cdef unsigned long prec = aprec
        cdef pAdicFixedModElement ans

        if mpz_fits_slong_p(self.prime_pow.prime.value) == 0:
            raise NotImplementedError("the prime %s does not fit in a long" % self.prime_pow.prime)
        p = self.prime_pow.prime

        ans = self._new_c()
        sig_on()
        padicexp(ans.value, self.value, p, prec, self.prime_pow.pow_mpz_t_tmp(prec))
        sig_off()

        return ans

    def _exp_newton(self, aprec, log_algorithm=None):
        r"""
        Compute the exponential power series of this element.

        This is a helper method for :meth:`exp`.

        INPUT:

        - ``aprec`` -- integer; the precision to which to compute the
          exponential

        - ``log_algorithm`` -- (default: ``None``) the algorithm used for
          computing the logarithm. This attribute is passed to the log
          method. See :meth:`log` for more details about the possible
          algorithms.

        .. NOTE::

            The function does not check that its argument ``self`` is
            the disk of convergence of ``exp``. If this assumption is not
            fulfilled the behaviour of the function is not specified.

        ALGORITHM:

        Solve the equation `\log(x) = self` using the Newton scheme::

        .. MATH::

            x_{i+1} = x_i \cdot (1 + self - \log(x_i))

        The binary complexity of this algorithm is roughly the same
        than that of the computation of the logarithm.

        EXAMPLES::

            sage: R.<w> = Zq(7^2,5)
            sage: x = R(7*w)
            sage: x.exp(algorithm='newton')   # indirect doctest
            1 + w*7 + (4*w + 2)*7^2 + (w + 6)*7^3 + 5*7^4 + O(7^5)
        """
        cdef unsigned long p
        cdef unsigned long prec = aprec
        cdef pAdicFixedModElement ans

        if mpz_fits_slong_p(self.prime_pow.prime.value) == 0:
            raise NotImplementedError("the prime %s does not fit in a long" % self.prime_pow.prime)
        p = self.prime_pow.prime

        ans = self._new_c()
        mpz_set_ui(ans.value, 1)
        sig_on()
        if p == 2:
            padicexp_Newton(ans.value, self.value, p, prec, 2, self.prime_pow.pow_mpz_t_tmp(prec))
        else:
            padicexp_Newton(ans.value, self.value, p, prec, 1, self.prime_pow.pow_mpz_t_tmp(prec))
        sig_off()

        return ans


def make_pAdicFixedModElement(parent, value):
    """
    Unpickles a fixed modulus element.

    EXAMPLES::

        sage: from sage.rings.padics.padic_fixed_mod_element import make_pAdicFixedModElement
        sage: R = ZpFM(5)
        sage: a = make_pAdicFixedModElement(R, 17*25); a
        2*5^2 + 3*5^3
    """
    return unpickle_fme_v2(pAdicFixedModElement, parent, value)



r"""
Supersingular Isogeny Diffie-Hellman

Toy implementation of the historical SIDH scheme.
This implementation serves as an example of a class that implements :class:`KeyExchangeBase`
rather than :class:`CommutativeKeyExchangeBase`.

.. WARNING::

    This is a toy implementation of a broken cryptographic scheme for educational
    use only! Do not use this implementation, or any cryptographic features of
    Sage, in any setting where security is needed!

AUTHORS:

- Taha Hedayat (2025-12-09): initial version
- Vincent Macri (2026-01-13): parameter set generation and cleanup
"""

# ****************************************************************************
#       Copyright (C) 2025 Taha Hedayat <tahah22121001@gmail.com>
#       Copyright (C) 2026 Vincent Macri <vincent.macri@ucalgary.ca>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from __future__ import annotations

import random
from typing import TYPE_CHECKING, Self

from sage.rings.finite_rings.finite_field_constructor import FiniteField
from sage.rings.integer import Integer
from sage.rings.integer_ring import ZZ
from sage.schemes.elliptic_curves.constructor import EllipticCurve

from .key_exchange_base import KeyExchangeBase

if TYPE_CHECKING:
    from sage.rings.finite_rings.element_base import FinitePolyExtElement
    from sage.schemes.elliptic_curves.ell_finite_field import EllipticCurve_finite_field
    from sage.schemes.elliptic_curves.ell_point import EllipticCurvePoint_finite_field
    from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite

    PublicKeySIDH = tuple[
        EllipticCurve_finite_field,
        EllipticCurvePoint_finite_field,
        EllipticCurvePoint_finite_field,
    ]
    SecretKeySIDH = Integer | int


class SIDH(KeyExchangeBase):
    r"""
    Supersingular isogeny Diffie-Hellman key exchange.

    This implementation uses the notation of [Cos2020]_.

    For an easier-to-use constructor, see :meth:`parameter_set_from_prime`
    which implements parameter set generation as described in [Jao2022]_.

    EXAMPLES:

    This example comes from [Cos2020]_.::

        sage: e_A = 4
        sage: e_B = 3
        sage: p = 2^e_A * 3^e_B - 1
        sage: K.<i> = GF(p^2, modulus=x^2 + 1)
        sage: a0 = 329 * i + 423
        sage: E = EllipticCurve(K, [0, a0, 0, 1, 0])
        sage: PA = E(100 * i + 248, 304 * i + 199)
        sage: QA = E(426 * i + 394, 51 * i + 79)
        sage: PB = E(358 * i + 275, 410 * i + 104)
        sage: QB = E(20 * i + 185, 281 * i + 239)
        sage: toy_sidh = key_exchange.SIDH(E, PA, QA, PB, QB)
        doctest:...: FutureWarning: SageMath's key exchange functionality is experimental and might change in the future.
                     See https://github.com/sagemath/sage/issues/41218 for details.
        sage: TestSuite(toy_sidh).run()
    """

    def __init__(
        self,
        E: EllipticCurve_finite_field,
        PA: EllipticCurvePoint_finite_field,
        QA: EllipticCurvePoint_finite_field,
        PB: EllipticCurvePoint_finite_field,
        QB: EllipticCurvePoint_finite_field,
    ) -> None:
        K = E.base_field()
        self._p: Integer = K.characteristic()
        n: Integer = self._p + 1
        self._e_A: Integer = n.valuation(2)
        self._e_B: Integer = n.valuation(3)

        if 2**self._e_A * 3**self._e_B - 1 != self._p:
            raise ValueError('curve must be over characteristic 2^e_A * 3^e_B - 1')

        if K.degree() != 2:
            raise ValueError('curve must be defined over a degree two field extension')

        self._E = E

        def validate_point(P):
            if P not in self._E:
                raise ValueError(f'{P} is not on {self._E}')

        validate_point(PA)
        validate_point(QA)
        validate_point(PB)
        validate_point(QB)
        self._PA = PA
        self._QA = QA
        self._PB = PB
        self._QB = QB

    @classmethod
    def parameter_set_from_prime(cls, prime: Integer | int) -> Self:
        r"""
        Generate an SIDH instance from the specified prime, generating the parameter
        set as in [Jao2022]_.

        EXAMPLES::

            sage: key_exchange.SIDH.parameter_set_from_prime(431)
            SIDH with parameter set: (431, Elliptic Curve defined by y^2 = x^3 + 6*x^2 + x over Finite Field in i of size 431^2, ...
        """
        p: Integer = Integer(prime)
        if not p.is_prime():
            raise ValueError('p must be prime')

        n: Integer = p + ZZ.one()
        e_A: Integer = n.valuation(2)
        e_B: Integer = n.valuation(3)

        x = ZZ['x'].gen()
        f = x**3 + 6 * x**2 + x
        Fp = FiniteField(p)

        Fp2 = Fp.extension(modulus=x**2 + 1, name='i')
        i = Fp2.gen()

        E = EllipticCurve(Fp2, [0, 6, 0, 1, 0])

        # As defined in [Jao2022]_
        def canonical_square_root(s) -> FinitePolyExtElement:
            r: FinitePolyExtElement = Fp2(s).sqrt()
            alpha, beta = r.list()
            alpha = alpha.lift()
            beta = beta.lift()
            if (alpha != 0 and alpha % 2 == 0) or (alpha == 0 and beta % 2 == 0):
                assert r**2 == s
                return r
            assert (-r) ** 2 == s
            return -r

        def trial_points():
            for c in range(p):
                x0 = Fp2(i + c)
                y0 = Fp2(f.subs(x=x0))
                if not y0.is_square():
                    continue
                y0 = canonical_square_root(y0)
                if E.is_on_curve(x0, y0):
                    yield E(x0, y0)

        PA_mults = (
            E(-3 + 2 * canonical_square_root(2), 0),
            E(-3 - 2 * canonical_square_root(2), 0),
        )
        for R in trial_points():
            PA = (3**e_B) * R
            if 2 ** (e_A - 1) * PA in PA_mults:
                break
        else:
            raise ValueError(
                'failed to find PA, try using SIDH constructor directly and specify all parameters'
            )

        E_00 = E(0, 0)
        for R in trial_points():
            QA = (3**e_B) * R
            if 2 ** (e_A - 1) * QA == E_00:
                break
        else:
            raise ValueError(
                'failed to find QA, try using SIDH constructor directly and specify all parameters'
            )

        for c in range(p):
            fc = f.subs(x=c)
            if not Fp(fc).is_square():
                continue
            PB = 2 ** (e_A - 1) * E(c, canonical_square_root(fc))
            if PB.order() == 3**e_B:
                break
        else:
            raise ValueError(
                'failed to find PB, try using SIDH constructor directly and specify all parameters'
            )

        for c in range(p):
            fc = f.subs(x=c)
            if Fp(fc).is_square() or not Fp2(fc).is_square():
                continue
            QB = 2 ** (e_A - 1) * E(c, canonical_square_root(fc))
            if QB.order() == 3**e_B:
                break
        else:
            raise ValueError(
                'failed to find QB, try using SIDH constructor directly and specify all parameters'
            )
        return cls(E, PA, QA, PB, QB)

    @classmethod
    def named_parameter_set(cls, name: str) -> Self:
        r"""
        Return an SIDH instance corresponding to a named parameter set.

        INPUT:

        - ``name`` -- one of the following:

            - ``"toy"``: a small SIDH instance over `\GF{431^2}`,
                         this toy example is from [Cos2020]_.
            - ``"p434"``: SIDH instance with the same parameters
                          as SIKEp434 from [Jao2022]_.
            - ``"p503"``: SIDH instance with the same parameters
                          as SIKEp503 from [Jao2022]_.
            - ``"p610"``: SIDH instance with the same parameters
                          as SIKEp610 from [Jao2022]_.
            - ``"p751"``: SIDH instance with the same parameters
                          as SIKEp751 from [Jao2022]_.

        TESTS:

        Test that p434 gives the same public basis points as the SIKE specification::

            sage: p434 = key_exchange.SIDH.named_parameter_set('p434')
            sage: p, E, P2, Q2, P3, Q3 = p434.parameters()
            sage: xQ20, xQ21 = Q2[0].list()
            sage: xQ20 == 0x0000C7461738340EFCF09CE388F666EB38F7F3AFD42DC0B664D9F461F31AA2EDC6B4AB71BD42F4D7C058E13F64B237EF7DDD2ABC0DEB0C6C
            True
            sage: xQ21 == 0x000025DE37157F50D75D320DD0682AB4A67E471586FBC2D31AA32E6957FA2B2614C4CD40A1E27283EAAF4272AE517847197432E2D61C85F5
            True
            sage: yQ20, yQ21 = Q2[1].list()
            sage: yQ20 == 0x0001D407B70B01E4AEE172EDF491F4EF32144F03F5E054CEF9FDE5A35EFA3642A11817905ED0D4F193F31124264924A5F64EFE14B6EC97E5
            True
            sage: yQ21 == 0x0000E7DEC8C32F50A4E735A839DCDB89FE0763A184C525F7B7D0EBC0E84E9D83E9AC53A572A25D19E1464B509D97272AE761657B4765B3D6
            True
            sage: xP20, xP21 = P2[0].list()
            sage: xP20 == 0x00003CCFC5E1F050030363E6920A0F7A4C6C71E63DE63A0E6475AF621995705F7C84500CB2BB61E950E19EAB8661D25C4A50ED279646CB48
            True
            sage: xP21 == 0x0001AD1C1CAE7840EDDA6D8A924520F60E573D3B9DFAC6D189941CB22326D284A8816CC4249410FE80D68047D823C97D705246F869E3EA50
            True
            sage: yP20, yP21 = P2[1].list()
            sage: yP20 == 0x0001AB066B84949582E3F66688452B9255E72A017C45B148D719D9A63CDB7BE6F48C812E33B68161D5AB3A0A36906F04A6A6957E6F4FB2E0
            True
            sage: yP21 == 0x0000FD87F67EA576CE97FF65BF9F4F7688C4C752DCE9F8BD2B36AD66E04249AAF8337C01E6E4E1A844267BA1A1887B433729E1DD90C7DD2F
            True
            sage: xQ30, xQ31 = Q3[0].list()
            sage: xQ30 == 0x00012E84D7652558E694BF84C1FBDAAF99B83B4266C32EC65B10457BCAF94C63EB063681E8B1E7398C0B241C19B9665FDB9E1406DA3D3846
            True
            sage: xQ31 == 0
            True
            sage: yQ30, yQ31 = Q3[1].list()
            sage: yQ30 == 0
            True
            sage: yQ31 == 0x0000EBAAA6C731271673BEECE467FD5ED9CC29AB564BDED7BDEAA86DD1E0FDDF399EDCC9B49C829EF53C7D7A35C3A0745D73C424FB4A5FD2
            True
            sage: xP30, xP31 = P3[0].list()
            sage: xP30 == 0x00008664865EA7D816F03B31E223C26D406A2C6CD0C3D667466056AAE85895EC37368BFC009DFAFCB3D97E639F65E9E45F46573B0637B7A9
            True
            sage: xP31 == 0
            True
            sage: yP30, yP31 = P3[1].list()
            sage: yP30 == 0x00006AE515593E73976091978DFBD70BDA0DD6BCAEEBFDD4FB1E748DDD9ED3FDCF679726C67A3B2CC12B39805B32B612E058A4280764443B
            True
            sage: yP31 == 0
            True
        """
        if name == 'toy':
            R = ZZ['x']
            x = R.gen()
            p = 2**4 * 3**3 - 1
            K = FiniteField(p**2, name='i', modulus=x**2 + 1)
            i = K.gen()
            E = EllipticCurve(K, [0, 329 * i + 423, 0, 1, 0])
            PA = E(100 * i + 248, 304 * i + 199)
            QA = E(426 * i + 394, 51 * i + 79)
            PB = E(358 * i + 275, 410 * i + 104)
            QB = E(20 * i + 185, 281 * i + 239)
            toy = cls(E, PA, QA, PB, QB)
            toy.rename('sidh-toy')
            return toy
        if name == 'p434':
            e_A = 0xD8
            e_B = 0x89
        elif name == 'p503':
            e_A = 0xFA
            e_B = 0x9F
        elif name == 'p610':
            e_A = 0x131
            e_B = 0xC0
        elif name == 'p751':
            e_A = 0x174
            e_B = 0xEF
        else:
            return super().named_parameter_set()

        p = 2**e_A * 3**e_B - 1
        sidh = cls.parameter_set_from_prime(p)
        sidh.rename(f'sidh-{name}')
        return sidh

    def parameters(
        self,
    ) -> tuple[
        Integer,
        EllipticCurve_finite_field,
        EllipticCurvePoint_finite_field,
        EllipticCurvePoint_finite_field,
        EllipticCurvePoint_finite_field,
        EllipticCurvePoint_finite_field,
    ]:
        r"""
        Return the parameter set of the SIDH instance.

        OUTPUT:

        A tuple (`E`, `PA`, `QA`, `PB`, `QB`) where:

        - `p` is the characteristic of the finite field `\GF{p^2}`
        - `E` is the starting curve
        - `PA` and `QA` are the generators for Alice's secret key
        - `PB` and `QB` are the generators for Bob's secret key
        """
        return (self._p, self._E, self._PA, self._QA, self._PB, self._QB)

    def alice_secret_key(self) -> Integer:
        r"""
        Generate Alice's secret key.

        EXAMPLES::

            sage: toy_sidh = key_exchange.SIDH.named_parameter_set('toy')
            sage: 0 <= toy_sidh.alice_secret_key() <= 2^4 - 1
            True
        """
        return Integer(random.randint(0, 2**self._e_A - 1))

    def bob_secret_key(self) -> Integer:
        r"""
        Generate Bob's secret key.

        EXAMPLES::

            sage: toy_sidh = key_exchange.SIDH.named_parameter_set('toy')
            sage: 0 <= toy_sidh.bob_secret_key() <= 3^3 - 1
            True
        """
        return Integer(random.randint(0, 3**self._e_B - 1))

    def alice_public_key(self, alice_secret_key: SecretKeySIDH) -> PublicKeySIDH:
        r"""
        Generate a valid public key for Alice.

        INPUT:

        - ``alice_secret_key`` -- Alice's secret key that will be used to generate
            the public key

        OUTPUT:

        Alice's public key as a tuple `(E_A, P'_B, Q'_B)`.
        """
        phi_A, _ = self.secret_isogeny_path(
            self._E, alice_secret_key, self._PA, self._QA
        )
        return phi_A.codomain(), phi_A(self._PB), phi_A(self._QB)

    def bob_public_key(self, bob_secret_key: SecretKeySIDH) -> PublicKeySIDH:
        r"""
        Generate a valid public key for Alice.

        INPUT:

        - ``alice_secret_key`` -- Alice's secret key that will be used to generate
            the public key

        OUTPUT:

        Bob's public key as a tuple `(E_B, P'_A, Q'_A)`.
        """
        phi_B, _ = self.secret_isogeny_path(self._E, bob_secret_key, self._PB, self._QB)
        return phi_B.codomain(), phi_B(self._PA), phi_B(self._QA)

    def alice_compute_shared_secret(
        self, alice_secret_key: SecretKeySIDH, bob_public_key: PublicKeySIDH
    ) -> Integer:
        r"""
        Compute the shared secret using Alice's secret key and Bob's public key.
        """
        E_B, P_A1, Q_A1 = bob_public_key
        phi_A1, _ = self.secret_isogeny_path(E_B, alice_secret_key, P_A1, Q_A1)
        E_AB = phi_A1.codomain()
        return E_AB.j_invariant()

    def bob_compute_shared_secret(
        self, bob_secret_key: SecretKeySIDH, alice_public_key: PublicKeySIDH
    ) -> Integer:
        r"""
        Compute the shared secret using Bob's secret key and Alice's public key.
        """
        E_A, P_B1, Q_B1 = alice_public_key
        phi_B1, _ = self.secret_isogeny_path(E_A, bob_secret_key, P_B1, Q_B1)
        E_BA = phi_B1.codomain()
        return E_BA.j_invariant()

    def secret_isogeny_path(
        self,
        start_curve: EllipticCurve_finite_field,
        secret_key: SecretKeySIDH,
        P: EllipticCurvePoint_finite_field,
        Q: EllipticCurvePoint_finite_field,
    ) -> tuple[EllipticCurveHom_composite, list[Integer]]:
        r"""
        Compute the secret isogeny and path of j-invariants.
        The kernel of returned isogeny is generated by ``P + secret_key * Q``.

        The purpose of this method is to demonstrate how implementations of
        :class:`~sage.crypto.public_key.key_exchange_base.KeyExchangeBase` may
        implement additional methods besides those defined in
        :class:`~sage.crypto.public_key.key_exchange_base.KeyExchangeBase`
        that compute some additional interesting information, in this case
        the path of j-invariants.

        INPUT:

        - ``start_curve`` -- the starting curve
        - ``secret_key`` -- the secret key
        - ``P`` -- first basis point
        - ``Q`` -- second basis point

        OUTPUT:

        A tuple ``(iso, path)`` where ``iso`` is an elliptic curve isogeny
        with kernel generated by ``P + secret_key * Q`` and ``path`` is
        a list of the j-invariants of the isomorphism classes of the curves on
        the isogeny path.

        EXAMPLES:

        We work through the key generation example from [Cos2020]_::

            sage: toy_sidh = key_exchange.SIDH.named_parameter_set('toy')
            sage: p, E0, PA, QA, PB, QB = toy_sidh.parameters()
            sage: alice_secret_key = 11
            sage: phi_A, alice_keygen_path = toy_sidh.secret_isogeny_path(E0, alice_secret_key, PA, QA)
            sage: alice_keygen_path
            [87*i + 190, 107, 344*i + 190, 350*i + 65, 222*i + 118]
            sage: bob_secret_key = 2
            sage: phi_B, bob_keygen_path = toy_sidh.secret_isogeny_path(E0, bob_secret_key, PB, QB)
            sage: bob_keygen_path
            [87*i + 190, 106*i + 379, 325*i + 379, 344*i + 190]
        """
        iso = start_curve.isogeny(P + secret_key * Q, algorithm='factored')
        j_invariant_path = [start_curve.j_invariant()]
        j_invariant_path.extend(E.codomain().j_invariant() for E in iso.factors())
        return iso, j_invariant_path



"""
Plane quartic curves over a general ring

These are generic genus 3 curves, as distinct from hyperelliptic curves of genus 3.

EXAMPLES::

    sage: PP.<X,Y,Z> = ProjectiveSpace(2, QQ)
    sage: f = X^4 + Y^4 + Z^4 - 3*X*Y*Z*(X+Y+Z)
    sage: C = QuarticCurve(f); C
    Quartic Curve over Rational Field
     defined by X^4 + Y^4 - 3*X^2*Y*Z - 3*X*Y^2*Z - 3*X*Y*Z^2 + Z^4
"""

# ***************************************************************************
#  Copyright (C) 2006 David Kohel <kohel@maths.usyd.edu>
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ***************************************************************************


from sage.schemes.curves import projective_curve


class QuarticCurve_generic(projective_curve.ProjectivePlaneCurve):
    # DRK: Note that we should check whether the curve is

    def _repr_type(self):
        """
        Return the representation of ``self``.

        EXAMPLES::

            sage: x,y,z = PolynomialRing(QQ, ['x','y','z']).gens()
            sage: Q = QuarticCurve(x**4 + y**4 + z**4)
            sage: Q._repr_type()
            'Quartic'
        """
        return "Quartic"

    def genus(self):
        """
        Return the genus of ``self``.

        EXAMPLES::

            sage: x,y,z = PolynomialRing(QQ, ['x','y','z']).gens()
            sage: Q = QuarticCurve(x**4 + y**4 + z**4)
            sage: Q.genus()
            3
        """
        return 3



# distutils: libraries = flint
# distutils: depends = flint/fq_poly_factor.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void fq_poly_factor_init(fq_poly_factor_t fac, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_clear(fq_poly_factor_t fac, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_realloc(fq_poly_factor_t fac, slong alloc, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_fit_length(fq_poly_factor_t fac, slong len, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_set(fq_poly_factor_t res, const fq_poly_factor_t fac, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_print_pretty(const fq_poly_factor_t fac, const char * var, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_print(const fq_poly_factor_t fac, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_insert(fq_poly_factor_t fac, const fq_poly_t poly, slong exp, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_concat(fq_poly_factor_t res, const fq_poly_factor_t fac, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_pow(fq_poly_factor_t fac, slong exp, const fq_ctx_t ctx) noexcept
    ulong fq_poly_remove(fq_poly_t f, const fq_poly_t p, const fq_ctx_t ctx) noexcept
    bint fq_poly_is_irreducible(const fq_poly_t f, const fq_ctx_t ctx) noexcept
    bint fq_poly_is_irreducible_ddf(const fq_poly_t f, const fq_ctx_t ctx) noexcept
    bint fq_poly_is_irreducible_ben_or(const fq_poly_t f, const fq_ctx_t ctx) noexcept
    bint _fq_poly_is_squarefree(const fq_struct * f, slong len, const fq_ctx_t ctx) noexcept
    bint fq_poly_is_squarefree(const fq_poly_t f, const fq_ctx_t ctx) noexcept
    bint fq_poly_factor_equal_deg_prob(fq_poly_t factor, flint_rand_t state, const fq_poly_t pol, slong d, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_equal_deg(fq_poly_factor_t factors, const fq_poly_t pol, slong d, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_split_single(fq_poly_t linfactor, const fq_poly_t input, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_distinct_deg(fq_poly_factor_t res, const fq_poly_t poly, slong * const * degs, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_squarefree(fq_poly_factor_t res, const fq_poly_t f, const fq_ctx_t ctx) noexcept
    void fq_poly_factor(fq_poly_factor_t res, fq_t lead, const fq_poly_t f, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_cantor_zassenhaus(fq_poly_factor_t res, const fq_poly_t f, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_kaltofen_shoup(fq_poly_factor_t res, const fq_poly_t poly, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_berlekamp(fq_poly_factor_t factors, const fq_poly_t f, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_with_berlekamp(fq_poly_factor_t res, fq_t leading_coeff, const fq_poly_t f, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_with_cantor_zassenhaus(fq_poly_factor_t res, fq_t leading_coeff, const fq_poly_t f, const fq_ctx_t ctx) noexcept
    void fq_poly_factor_with_kaltofen_shoup(fq_poly_factor_t res, fq_t leading_coeff, const fq_poly_t f, const fq_ctx_t ctx) noexcept
    void fq_poly_iterated_frobenius_preinv(fq_poly_t * rop, slong n, const fq_poly_t v, const fq_poly_t vinv, const fq_ctx_t ctx) noexcept
    void fq_poly_roots(fq_poly_factor_t r, const fq_poly_t f, int with_multiplicity, const fq_ctx_t ctx) noexcept



# distutils: libraries = GSL_LIBRARIES
# distutils: library_dirs = GSL_LIBDIR
# distutils: include_dirs = GSL_INCDIR
from sage.libs.gsl.types cimport *

cdef extern from "gsl/gsl_randist.h":
  unsigned int gsl_ran_bernoulli ( gsl_rng * r, double p)
  double gsl_ran_bernoulli_pdf ( unsigned int k, double p)

  double gsl_ran_beta ( gsl_rng * r,  double a,  double b)
  double gsl_ran_beta_pdf ( double x,  double a,  double b)

  unsigned int gsl_ran_binomial ( gsl_rng * r, double p, unsigned int n)
  unsigned int gsl_ran_binomial_tpe ( gsl_rng * r, double pp, unsigned int n)
  double gsl_ran_binomial_pdf ( unsigned int k,  double p,  unsigned int n)

  double gsl_ran_exponential ( gsl_rng * r,  double mu)
  double gsl_ran_exponential_pdf ( double x,  double mu)

  double gsl_ran_exppow ( gsl_rng * r,  double a,  double b)
  double gsl_ran_exppow_pdf ( double x,  double a,  double b)

  double gsl_ran_cauchy ( gsl_rng * r,  double a)
  double gsl_ran_cauchy_pdf ( double x,  double a)

  double gsl_ran_chisq ( gsl_rng * r,  double nu)
  double gsl_ran_chisq_pdf ( double x,  double nu)

  void gsl_ran_dirichlet ( gsl_rng * r,  size_t K,  double alpha[], double theta[])
  double gsl_ran_dirichlet_pdf ( size_t K,  double alpha[],  double theta[])
  double gsl_ran_dirichlet_lnpdf ( size_t K,  double alpha[],  double theta[])

  double gsl_ran_erlang ( gsl_rng * r,  double a,  double n)
  double gsl_ran_erlang_pdf ( double x,  double a,  double n)

  double gsl_ran_fdist ( gsl_rng * r,  double nu1,  double nu2)
  double gsl_ran_fdist_pdf ( double x,  double nu1,  double nu2)

  double gsl_ran_flat ( gsl_rng * r,  double a,  double b)
  double gsl_ran_flat_pdf (double x,  double a,  double b)

  double gsl_ran_gamma ( gsl_rng * r,  double a,  double b)
  double gsl_ran_gamma_int ( gsl_rng * r,  unsigned int a)
  double gsl_ran_gamma_pdf ( double x,  double a,  double b)

  double gsl_ran_gaussian ( gsl_rng * r,  double sigma)
  double gsl_ran_gaussian_ratio_method ( gsl_rng * r,  double sigma)
  double gsl_ran_gaussian_pdf ( double x,  double sigma)

  double gsl_ran_ugaussian ( gsl_rng * r)
  double gsl_ran_ugaussian_ratio_method ( gsl_rng * r)
  double gsl_ran_ugaussian_pdf ( double x)

  double gsl_ran_gaussian_tail ( gsl_rng * r,  double a,  double sigma)
  double gsl_ran_gaussian_tail_pdf ( double x,  double a,  double sigma)

  double gsl_ran_ugaussian_tail ( gsl_rng * r,  double a)
  double gsl_ran_ugaussian_tail_pdf ( double x,  double a)

  void gsl_ran_bivariate_gaussian ( gsl_rng * r, double sigma_x, double sigma_y, double rho, double *x, double *y)
  double gsl_ran_bivariate_gaussian_pdf ( double x,  double y,  double sigma_x,  double sigma_y,  double rho)

  double gsl_ran_landau ( gsl_rng * r)
  double gsl_ran_landau_pdf ( double x)

  unsigned int gsl_ran_geometric ( gsl_rng * r,  double p)
  double gsl_ran_geometric_pdf ( unsigned int k,  double p)

  unsigned int gsl_ran_hypergeometric ( gsl_rng * r, unsigned int n1, unsigned int n2, unsigned int t)
  double gsl_ran_hypergeometric_pdf ( unsigned int k,  unsigned int n1,  unsigned int n2, unsigned int t)

  double gsl_ran_gumbel1 ( gsl_rng * r,  double a,  double b)
  double gsl_ran_gumbel1_pdf ( double x,  double a,  double b)

  double gsl_ran_gumbel2 ( gsl_rng * r,  double a,  double b)
  double gsl_ran_gumbel2_pdf ( double x,  double a,  double b)

  double gsl_ran_logistic ( gsl_rng * r,  double a)
  double gsl_ran_logistic_pdf ( double x,  double a)

  double gsl_ran_lognormal ( gsl_rng * r,  double zeta,  double sigma)
  double gsl_ran_lognormal_pdf ( double x,  double zeta,  double sigma)

  unsigned int gsl_ran_logarithmic ( gsl_rng * r,  double p)
  double gsl_ran_logarithmic_pdf ( unsigned int k,  double p)

  void gsl_ran_multinomial ( gsl_rng * r,  size_t K,
                             unsigned int N,  double p[],
                            unsigned int n[] )
  double gsl_ran_multinomial_pdf ( size_t K,
                                   double p[],  unsigned int n[] )
  double gsl_ran_multinomial_lnpdf ( size_t K,
                              double p[],  unsigned int n[] )


  unsigned int gsl_ran_negative_binomial ( gsl_rng * r, double p, double n)
  double gsl_ran_negative_binomial_pdf ( unsigned int k,  double p, double n)

  unsigned int gsl_ran_pascal ( gsl_rng * r, double p, unsigned int n)
  double gsl_ran_pascal_pdf ( unsigned int k,  double p, unsigned int n)

  double gsl_ran_pareto ( gsl_rng * r, double a,  double b)
  double gsl_ran_pareto_pdf ( double x,  double a,  double b)

  unsigned int gsl_ran_poisson ( gsl_rng * r, double mu)
  void gsl_ran_poisson_array ( gsl_rng * r, size_t n, unsigned int array[],
                              double mu)
  double gsl_ran_poisson_pdf ( unsigned int k,  double mu)

  double gsl_ran_rayleigh ( gsl_rng * r,  double sigma)
  double gsl_ran_rayleigh_pdf ( double x,  double sigma)

  double gsl_ran_rayleigh_tail ( gsl_rng * r,  double a,  double sigma)
  double gsl_ran_rayleigh_tail_pdf ( double x,  double a,  double sigma)

  double gsl_ran_tdist ( gsl_rng * r,  double nu)
  double gsl_ran_tdist_pdf ( double x,  double nu)

  double gsl_ran_fdist ( gsl_rng * r,  double a,  double b)
  double gsl_ran_fdist_pdf ( double x,  double a,  double b)

  double gsl_ran_laplace ( gsl_rng * r,  double a)
  double gsl_ran_laplace_pdf ( double x,  double a)

  double gsl_ran_levy ( gsl_rng * r,  double c,  double alpha)
  double gsl_ran_levy_skew ( gsl_rng * r,  double c,  double alpha,  double beta)

  double gsl_ran_weibull ( gsl_rng * r,  double a,  double b)
  double gsl_ran_weibull_pdf ( double x,  double a,  double b)

  void gsl_ran_dir_2d ( gsl_rng * r, double * x, double * y)
  void gsl_ran_dir_2d_trig_method ( gsl_rng * r, double * x, double * y)
  void gsl_ran_dir_3d ( gsl_rng * r, double * x, double * y, double * z)
  void gsl_ran_dir_nd ( gsl_rng * r, size_t n, double * x)

  void gsl_ran_shuffle ( gsl_rng * r, void * base, size_t nmembm, size_t size)
  int gsl_ran_choose ( gsl_rng * r, void * dest, size_t k, void * src, size_t n, size_t size)
  void gsl_ran_sample ( gsl_rng * r, void * dest, size_t k, void * src, size_t n, size_t size)


  ctypedef struct gsl_ran_discrete_t

  gsl_ran_discrete_t * gsl_ran_discrete_preproc (size_t K,  double *P)
  void gsl_ran_discrete_free(gsl_ran_discrete_t *g)
  size_t gsl_ran_discrete ( gsl_rng *r,  gsl_ran_discrete_t *g)
  double gsl_ran_discrete_pdf (size_t k,  gsl_ran_discrete_t *g)

cdef extern from "gsl/gsl_cdf.h":
  double gsl_cdf_ugaussian_P ( double x)
  double gsl_cdf_ugaussian_Q ( double x)

  double gsl_cdf_ugaussian_Pinv ( double P)
  double gsl_cdf_ugaussian_Qinv ( double Q)

  double gsl_cdf_gaussian_P ( double x,  double sigma)
  double gsl_cdf_gaussian_Q ( double x,  double sigma)

  double gsl_cdf_gaussian_Pinv ( double P,  double sigma)
  double gsl_cdf_gaussian_Qinv ( double Q,  double sigma)

  double gsl_cdf_gamma_P ( double x,  double a,  double b)
  double gsl_cdf_gamma_Q ( double x,  double a,  double b)

  double gsl_cdf_gamma_Pinv ( double P,  double a,  double b)
  double gsl_cdf_gamma_Qinv ( double Q,  double a,  double b)

  double gsl_cdf_cauchy_P ( double x,  double a)
  double gsl_cdf_cauchy_Q ( double x,  double a)

  double gsl_cdf_cauchy_Pinv ( double P,  double a)
  double gsl_cdf_cauchy_Qinv ( double Q,  double a)

  double gsl_cdf_laplace_P ( double x,  double a)
  double gsl_cdf_laplace_Q ( double x,  double a)

  double gsl_cdf_laplace_Pinv ( double P,  double a)
  double gsl_cdf_laplace_Qinv ( double Q,  double a)

  double gsl_cdf_rayleigh_P ( double x,  double sigma)
  double gsl_cdf_rayleigh_Q ( double x,  double sigma)

  double gsl_cdf_rayleigh_Pinv ( double P,  double sigma)
  double gsl_cdf_rayleigh_Qinv ( double Q,  double sigma)

  double gsl_cdf_chisq_P ( double x,  double nu)
  double gsl_cdf_chisq_Q ( double x,  double nu)

  double gsl_cdf_chisq_Pinv ( double P,  double nu)
  double gsl_cdf_chisq_Qinv ( double Q,  double nu)

  double gsl_cdf_exponential_P ( double x,  double mu)
  double gsl_cdf_exponential_Q ( double x,  double mu)

  double gsl_cdf_exppow_P(double x, double a,double b)

  double gsl_cdf_exponential_Pinv ( double P,  double mu)
  double gsl_cdf_exponential_Qinv ( double Q,  double mu)

  double gsl_cdf_tdist_P ( double x,  double nu)
  double gsl_cdf_tdist_Q ( double x,  double nu)

  double gsl_cdf_tdist_Pinv ( double P,  double nu)
  double gsl_cdf_tdist_Qinv ( double Q,  double nu)

  double gsl_cdf_fdist_P ( double x,  double nu1,  double nu2)
  double gsl_cdf_fdist_Q ( double x,  double nu1,  double nu2)

  double gsl_cdf_fdist_Pinv(double x, double a, double b)
  double gsl_cdf_fdist_Qinv(double x, double a, double b)

  double gsl_cdf_beta_P ( double x,  double a,  double b)
  double gsl_cdf_beta_Q ( double x,  double a,  double b)

  double gsl_cdf_beta_Pinv(double x, double a, double b)

  double gsl_cdf_flat_P ( double x,  double a,  double b)
  double gsl_cdf_flat_Q ( double x,  double a,  double b)

  double gsl_cdf_flat_Pinv ( double P,  double a,  double b)
  double gsl_cdf_flat_Qinv ( double Q,  double a,  double b)

  double gsl_cdf_lognormal_P ( double x,  double zeta,  double sigma)
  double gsl_cdf_lognormal_Q ( double x,  double zeta,  double sigma)

  double gsl_cdf_lognormal_Pinv ( double P,  double zeta,  double sigma)
  double gsl_cdf_lognormal_Qinv ( double Q,  double zeta,  double sigma)

  double gsl_cdf_gumbel1_P ( double x,  double a,  double b)
  double gsl_cdf_gumbel1_Q ( double x,  double a,  double b)

  double gsl_cdf_gumbel1_Pinv ( double P,  double a,  double b)
  double gsl_cdf_gumbel1_Qinv ( double Q,  double a,  double b)

  double gsl_cdf_gumbel2_P ( double x,  double a,  double b)
  double gsl_cdf_gumbel2_Q ( double x,  double a,  double b)

  double gsl_cdf_gumbel2_Pinv ( double P,  double a,  double b)
  double gsl_cdf_gumbel2_Qinv ( double Q,  double a,  double b)

  double gsl_cdf_weibull_P ( double x,  double a,  double b)
  double gsl_cdf_weibull_Q ( double x,  double a,  double b)

  double gsl_cdf_weibull_Pinv ( double P,  double a,  double b)
  double gsl_cdf_weibull_Qinv ( double Q,  double a,  double b)

  double gsl_cdf_pareto_P ( double x,  double a,  double b)
  double gsl_cdf_pareto_Q ( double x,  double a,  double b)

  double gsl_cdf_pareto_Pinv ( double P,  double a,  double b)
  double gsl_cdf_pareto_Qinv ( double Q,  double a,  double b)

  double gsl_cdf_logistic_P ( double x,  double a)
  double gsl_cdf_logistic_Q ( double x,  double a)

  double gsl_cdf_logistic_Pinv ( double P,  double a)
  double gsl_cdf_logistic_Qinv ( double Q,  double a)



r"""
Index of code constructions

The ``codes`` object may be used to access the codes that Sage can build.

Families of Codes (Rich representation)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: @

    :meth:`~sage.coding.parity_check_code.ParityCheckCode` @ Parity check codes
    :meth:`~sage.coding.cyclic_code.CyclicCode` @ Cyclic codes
    :meth:`~sage.coding.bch_code.BCHCode` @ BCH Codes
    :meth:`~sage.coding.grs_code.GeneralizedReedSolomonCode` @ Generalized Reed-Solomon codes
    :meth:`~sage.coding.grs_code.ReedSolomonCode` @ Reed-Solomon codes
    :meth:`~sage.coding.reed_muller_code.BinaryReedMullerCode` @ Binary Reed-Muller codes
    :meth:`~sage.coding.reed_muller_code.ReedMullerCode` @ q-ary Reed-Muller codes
    :meth:`~sage.coding.hamming_code.HammingCode` @ Hamming codes
    :meth:`~sage.coding.golay_code.GolayCode` @ Golay codes
    :meth:`~sage.coding.goppa_code.GoppaCode` @ Goppa codes
    :meth:`~sage.coding.kasami_codes.KasamiCode` @ Kasami codes

Families of Codes (Generator matrix representation)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: @

    :meth:`~sage.coding.code_constructions.DuadicCodeEvenPair` @ Duadic codes, even pair
    :meth:`~sage.coding.code_constructions.DuadicCodeOddPair` @ Duadic codes, odd pair
    :meth:`~sage.coding.code_constructions.QuadraticResidueCode` @  Quadratic residue codes
    :meth:`~sage.coding.code_constructions.ExtendedQuadraticResidueCode` @ Extended quadratic residue codes
    :meth:`~sage.coding.code_constructions.QuadraticResidueCodeEvenPair` @ Even-like quadratic residue codes
    :meth:`~sage.coding.code_constructions.QuadraticResidueCodeOddPair` @ Odd-like quadratic residue codes
    :meth:`~sage.coding.guava.QuasiQuadraticResidueCode` @ Quasi quadratic residue codes (Requires GAP/Guava)
    :meth:`~sage.coding.code_constructions.ToricCode` @ Toric codes
    :meth:`~sage.coding.code_constructions.WalshCode` @ Walsh codes
    :meth:`~sage.coding.code_constructions.from_parity_check_matrix` @ Construct a code from a parity check matrix
    :meth:`~sage.coding.code_constructions.random_linear_code` @ Construct a random linear code
    :meth:`~sage.coding.guava.RandomLinearCodeGuava` @ Construct a random linear code through Guava (Requires GAP/Guava)


Derived Codes
^^^^^^^^^^^^^

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: @

    :meth:`~sage.coding.subfield_subcode.SubfieldSubcode` @ Subfield subcodes
    :meth:`~sage.coding.extended_code.ExtendedCode` @ Extended codes
    :meth:`~sage.coding.punctured_code.PuncturedCode` @ Puncturedcodes

To import these names into the global namespace, use::

    sage: from sage.coding.codes_catalog import *
"""
#*****************************************************************************
#       Copyright (C) 2009 David Lucas <david.lucas@inria.fr>
#
#  Distributed under the terms of the GNU General Public License (GPL),
#  version 2 or later (at your preference).
#
#                  https://www.gnu.org/licenses/
#*****************************************************************************

# This module is imported as "codes" in all.py so that codes.<tab> is
# available in the global namespace.

from sage.misc.lazy_import import lazy_import as _lazy_import

from .linear_code import LinearCode

_lazy_import('sage.coding.code_constructions',
        ['DuadicCodeEvenPair', 'DuadicCodeOddPair',
         'ExtendedQuadraticResidueCode', 'from_parity_check_matrix',
         'QuadraticResidueCode', 'QuadraticResidueCodeEvenPair',
         'QuadraticResidueCodeOddPair', 'random_linear_code',
         'ToricCode', 'WalshCode'])

_lazy_import('sage.coding.subfield_subcode', 'SubfieldSubcode')
_lazy_import('sage.coding.extended_code', 'ExtendedCode')
_lazy_import('sage.coding.punctured_code', 'PuncturedCode')

_lazy_import('sage.coding.parity_check_code', 'ParityCheckCode')
_lazy_import('sage.coding.cyclic_code', 'CyclicCode')
_lazy_import('sage.coding.bch_code', 'BCHCode')
_lazy_import('sage.coding.grs_code', ['GeneralizedReedSolomonCode', 'ReedSolomonCode'])
_lazy_import('sage.coding.reed_muller_code', ['BinaryReedMullerCode', 'ReedMullerCode'])
_lazy_import('sage.coding.hamming_code', 'HammingCode')
_lazy_import('sage.coding.golay_code', 'GolayCode')
_lazy_import('sage.coding.goppa_code', 'GoppaCode')
_lazy_import('sage.coding.kasami_codes', 'KasamiCode')
_lazy_import('sage.coding.linear_rank_metric', 'LinearRankMetricCode')
_lazy_import('sage.coding.gabidulin_code', 'GabidulinCode')
_lazy_import('sage.coding.ag_code', ['EvaluationAGCode', 'DifferentialAGCode', 'CartierCode'])

_lazy_import('sage.coding.guava', ['QuasiQuadraticResidueCode', 'RandomLinearCodeGuava'])

from . import decoders_catalog as decoders
from . import encoders_catalog as encoders
from . import bounds_catalog as bounds

_lazy_import('sage.coding','databases')



#!/usr/bin/env python3

r"""
This postprocess script fixes some issues with .rst files returned by nbconvert:

- it fixes the rendering of the first title (that looks like markdown not rst)
- it removes the header that defines some LaTeX macros
- it removes the :math: role since this is the default one

AUTHORS:

- Thierry Monteil (2018): initial version.
"""

if __name__ == '__main__':

    import sys
    import re

    file_name = sys.argv[1]

    with open(file_name) as f:
        lines = f.readlines()

    # states of the parser
    wrong_math_def_started = False
    wrong_math_fixed = False
    wrong_title_fixed = False

    # processing
    new_file = ''
    for i, line in enumerate(lines):
        if line.startswith(' # ') and not wrong_title_fixed:
            new_file += re.sub('^ # ', '', line)
            new_file += '=' * (len(line) - 4) + '\n'
            wrong_title_fixed = True
        elif line.startswith('.. math::') and not wrong_math_fixed:
            pass
        elif line.startswith('   \\def') and not wrong_math_fixed:
            wrong_math_def_started = True
        elif wrong_math_def_started and not wrong_math_fixed:
            wrong_math_fixed = True
        else:
            new_file += re.sub(':math:', '', line)

    # write new file
    with open(file_name, 'w') as f:
        f.write(new_file)



r"""
Finite Field Diffie-Hellman

Toy implementation of Diffie-Hellman key exchange over finite fields `\Zmod{p}`.

.. WARNING::

    This is a toy implementation for educational use only! Do not use
    this implementation, or any cryptographic features of Sage, in any
    setting where security is needed!

AUTHORS:

- Vincent Macri (2024-07-30): initial version
- Brian Heckel (2025-12-06): converted to inherit from KeyExchangeBase class
"""

# ****************************************************************************
#       Copyright (C) 2024 Vincent Macri <vincent.macri@ucalgary.ca>
#       Copyright (C) 2025 Brian Heckel <heckelbri@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from __future__ import annotations

from typing import TYPE_CHECKING

from sage.crypto.public_key.key_exchange.key_exchange_base import (
    CommutativeKeyExchangeBase,
)
from sage.misc.prandom import randint
from sage.rings.finite_rings.finite_field_constructor import GF
from sage.rings.integer import Integer

if TYPE_CHECKING:
    from sage.rings.finite_rings.finite_field_prime_modn import FiniteField_prime_modn
    from sage.rings.finite_rings.integer_mod import IntegerMod_abstract


class FiniteFieldDiffieHellman(CommutativeKeyExchangeBase):
    def __init__(
        self,
        p: Integer | int,
        generator: Integer | IntegerMod_abstract | int,
        proof: bool | None = None,
    ) -> None:
        r"""
        Create an instance of the Diffie-Hellman key exchange scheme using the
        given prime ``p`` and base ``g``.

        INPUT:

        - ``p`` -- prime integer defining the field `\GF{p}` that the key
          exchanges will be performed over, must be at least 5

        - ``generator`` -- base for the key exchange, (coerceable to) an element of
          `\GF{p}` from `2` to `p - 2`

        - ``proof`` -- whether to require a proof that ``p`` is prime.
          If ``False``, a probabilistic test is used to check that ``p`` is
          prime. This should be set to ``False`` when using large
          (cryptographic size) primes, otherwise checking primality will take
          too long. If this is not specified, then the default behaviour is
          to use the current value of `proof.arithmetic()`.

        REFERENCES:

        For more information, see Section 8.1 of [PP2010]_.

        EXAMPLES::

            sage: DH = key_exchange.FiniteFieldDiffieHellman(13, 2)
            doctest:...: FutureWarning: SageMath's key exchange functionality is experimental and might change in the future.
                         See https://github.com/sagemath/sage/issues/41218 for details.

        This is an example of a full key exchange using a cryptographically
        large prime. This is the prime from the 8192-bit MODP group in RFC 3526
        (see [KK2003]_)::

            sage: p = 2^8192 - 2^8128 - 1 + 2^64 * (round(2^8062 * pi) + 4743158)
            sage: DH = key_exchange.FiniteFieldDiffieHellman(p, 2, proof=False)
            sage: alice_sk = DH.secret_key()
            sage: alice_pk = DH.public_key(alice_sk)
            sage: bob_sk = DH.secret_key()
            sage: bob_pk = DH.public_key(bob_sk)
            sage: alice_shared_secret = DH.compute_shared_secret(alice_sk, bob_pk)
            sage: bob_shared_secret = DH.compute_shared_secret(bob_sk, alice_pk)
            sage: alice_shared_secret == bob_shared_secret
            True

        TESTS::

            sage: DH = key_exchange.FiniteFieldDiffieHellman(5, 3)
            sage: TestSuite(DH).run()
        """
        if p < 5:
            raise ValueError('p must be at least 5')

        self._p = Integer(p)
        # The modn implementation checks that ``p`` is prime
        self._field = GF(self._p, impl='modn', proof=proof)

        self._generator = self._field(generator)

    def secret_key(self) -> IntegerMod_abstract:
        """
        Generate a random Diffie-Hellman secret key.

        TESTS:

            sage: DH = key_exchange.FiniteFieldDiffieHellman(7, 2)
            sage: keys = [DH.secret_key() for i in range(10)]
            sage: all(2 <= i <= 5 for i in keys)
            True
        """
        return self._field(randint(2, self._p - 2))

    def public_key(self, secret_key) -> IntegerMod_abstract:
        """
        Generate a Diffie-Hellman public key using the given secret key.

        INPUT:

        - ``secret_key`` -- the secret key to generate the public key with

        EXAMPLES::

            sage: DH = key_exchange.FiniteFieldDiffieHellman(13, 2)
            sage: DH.public_key(4)
            3
        """
        return self._generator**secret_key

    def compute_shared_secret(self, secret_key, public_key) -> IntegerMod_abstract:
        """
        Compute the shared secret using the given public key and secret keys.

        INPUT:

        - ``pk`` -- public key

        - ``sk`` -- secret key

        EXAMPLES::

            sage: DH = key_exchange.FiniteFieldDiffieHellman(17, 3)
            sage: DH.compute_shared_secret(11, 13)
            4
        """
        return self._field(public_key**secret_key)

    def parameters(self) -> tuple[FiniteField_prime_modn, IntegerMod_abstract]:
        """
        Returns a list of the prime ``p`` used, the base generator ``g`` and the field used.

        EXAMPLES::

            sage: DH = key_exchange.FiniteFieldDiffieHellman(17, 3)
            sage: DH.parameters()
            (Finite Field of size 17, 3)
        """
        return (self._field, self._generator)



# sage.doctest: needs sage.graphs
"""
Catalog of simplicial complexes

There are two main types: manifolds and examples related to graph
theory.

For manifolds, there are functions defining the `n`-sphere for any
`n`, the torus, `n`-dimensional real projective space for any `n`, the
complex projective plane, surfaces of arbitrary genus, and some other
manifolds, all as simplicial complexes.

Aside from surfaces, this file also provides functions for
constructing some other simplicial complexes: the simplicial complex
of not-`i`-connected graphs on `n` vertices, the matching complex on n
vertices, the chessboard complex for an `n` by `i` chessboard, and
others.  These provide examples of large simplicial complexes; for
example, ``simplicial_complexes.NotIConnectedGraphs(7,2)`` has over a
million simplices.

All of these examples are accessible by typing
``simplicial_complexes.NAME``, where ``NAME`` is the name of the example.

- :meth:`~sage.topology.examples.BarnetteSphere`
- :meth:`~sage.topology.examples.BrucknerGrunbaumSphere`
- :meth:`~sage.topology.examples.ChessboardComplex`
- :meth:`~sage.topology.examples.ComplexProjectivePlane`
- :meth:`~sage.topology.examples.DunceHat`
- :meth:`~sage.topology.examples.FareyMap`
- :meth:`~sage.topology.examples.GenusSix`
- :meth:`~sage.topology.examples.HochschildSphere`
- :meth:`~sage.topology.examples.K3Surface`
- :meth:`~sage.topology.examples.KleinBottle`
- :meth:`~sage.topology.examples.MatchingComplex`
- :meth:`~sage.topology.examples.MooreSpace`
- :meth:`~sage.topology.examples.NoncrossingBipartiteComplex`
- :meth:`~sage.topology.examples.NotIConnectedGraphs`
- :meth:`~sage.topology.examples.PoincareHomologyThreeSphere`
- :meth:`~sage.topology.examples.QuaternionicProjectivePlane`
- :meth:`~sage.topology.examples.RandomComplex`
- :meth:`~sage.topology.examples.RandomTwoSphere`
- :meth:`~sage.topology.examples.RealProjectivePlane`
- :meth:`~sage.topology.examples.RealProjectiveSpace`
- :meth:`~sage.topology.examples.RudinBall`
- :meth:`~sage.topology.examples.ShiftedComplex`
- :meth:`~sage.topology.examples.Simplex`
- :meth:`~sage.topology.examples.Sphere`
- :meth:`~sage.topology.examples.SumComplex`
- :meth:`~sage.topology.examples.SurfaceOfGenus`
- :meth:`~sage.topology.examples.Torus`
- :meth:`~sage.topology.examples.ZieglerBall`

You can also get a list by typing ``simplicial_complexes.`` and hitting the
:kbd:`Tab` key.

EXAMPLES::

    sage: S = simplicial_complexes.Sphere(2) # the 2-sphere
    sage: S.homology()                                                                  # needs sage.modules
    {0: 0, 1: 0, 2: Z}
    sage: simplicial_complexes.SurfaceOfGenus(3)
    Triangulation of an orientable surface of genus 3
    sage: M4 = simplicial_complexes.MooreSpace(4)
    sage: M4.homology()                                                                 # needs sage.modules
    {0: 0, 1: C4, 2: 0}
    sage: simplicial_complexes.MatchingComplex(6).homology()                            # needs sage.modules
    {0: 0, 1: Z^16, 2: 0}
"""

from sage.topology.simplicial_complex_examples import (Sphere, Simplex, Torus,
                                                       ProjectivePlane,
                                                       RealProjectivePlane, KleinBottle,
                                                       FareyMap, GenusSix,
                                                       SurfaceOfGenus,
                                                       MooreSpace,
                                                       ComplexProjectivePlane,
                                                       QuaternionicProjectivePlane,
                                                       PoincareHomologyThreeSphere,
                                                       RealProjectiveSpace, K3Surface,
                                                       BarnetteSphere,
                                                       BrucknerGrunbaumSphere,
                                                       NotIConnectedGraphs,
                                                       MatchingComplex,
                                                       ChessboardComplex, RandomComplex,
                                                       SumComplex,
                                                       RandomTwoSphere, ShiftedComplex,
                                                       RudinBall, ZieglerBall, DunceHat)

from sage.combinat.posets.hochschild_lattice import hochschild_simplicial_complex as HochschildSphere

from sage.combinat.posets.bubble_shuffle import noncrossing_bipartite_complex as NoncrossingBipartiteComplex



# sage.doctest: needs sage.groups
r"""
Testing arithmetic subgroup
"""
################################################################################
#
#       Copyright (C) 2009, The Sage Group -- http://www.sagemath.org/
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
#
################################################################################
from __future__ import annotations

from sage.misc import prandom
from sage.misc.timing import cputime
from sage.modular.arithgroup.arithgroup_perm import (
    ArithmeticSubgroup_Permutation,
    EvenArithmeticSubgroup_Permutation,
    OddArithmeticSubgroup_Permutation,
)
from sage.modular.arithgroup.congroup_gamma import Gamma_constructor as Gamma
from sage.modular.arithgroup.congroup_gamma0 import Gamma0_constructor as Gamma0
from sage.modular.arithgroup.congroup_gamma1 import Gamma1_constructor as Gamma1
from sage.modular.arithgroup.congroup_gammaH import GammaH_constructor as GammaH
from sage.rings.finite_rings.integer_mod_ring import Zmod


def random_even_arithgroup(index, nu2_max=None, nu3_max=None):
    r"""
    Return a random even arithmetic subgroup.

    EXAMPLES::

        sage: import sage.modular.arithgroup.tests as tests
        sage: G = tests.random_even_arithgroup(30); G # random
        Arithmetic subgroup of index 30
        sage: G.is_even()
        True
    """
    from sage.groups.perm_gps.permgroup import PermutationGroup

    test = False

    if nu2_max is None:
        nu2_max = index // 5
    elif nu2_max == 0:
        assert index % 2 == 0
    if nu3_max is None:
        nu3_max = index // 7
    elif nu3_max == 0:
        assert index % 3 == 0

    while not test:
        nu2 = prandom.randint(0, nu2_max)
        nu2 = index % 2 + nu2 * 2
        nu3 = prandom.randint(0, nu3_max)
        nu3 = index % 3 + nu3 * 3

        l = list(range(1, index + 1))
        prandom.shuffle(l)
        S2: list[tuple] = []
        for i in range(nu2):
            S2.append((l[i],))
        for i in range(nu2, index, 2):
            S2.append((l[i], l[i + 1]))
        prandom.shuffle(l)
        S3: list[tuple] = []
        for i in range(nu3):
            S3.append((l[i],))
        for i in range(nu3, index, 3):
            S3.append((l[i], l[i + 1], l[i + 2]))
        G = PermutationGroup([S2, S3])
        test = G.is_transitive()

    return ArithmeticSubgroup_Permutation(S2=S2, S3=S3)


def random_odd_arithgroup(index, nu3_max=None):
    r"""
    Return a random odd arithmetic subgroup.

    EXAMPLES::

        sage: from sage.modular.arithgroup.tests import random_odd_arithgroup
        sage: G = random_odd_arithgroup(20); G #random
        Arithmetic subgroup of index 20
        sage: G.is_odd()
        True
    """
    assert index % 4 == 0
    G = random_even_arithgroup(index // 2, nu2_max=0, nu3_max=nu3_max)
    return G.one_odd_subgroup(random=True)


class Test:
    r"""
    Testing class for arithmetic subgroup implemented via permutations.
    """

    def __init__(self, index=20, index_max=50, odd_probability=0.5):
        r"""
        Create an arithmetic subgroup testing object.

        INPUT:

        - ``index`` -- the index of random subgroup to test

        - ``index_max`` -- the maximum index for congruence subgroup to test

        EXAMPLES::

            sage: from sage.modular.arithgroup.tests import Test
            sage: Test()
            Arithmetic subgroup testing class
        """
        self.congroups = []
        i = 1
        self.odd_probability = odd_probability
        if index % 4:
            self.odd_probability = 0
        while Gamma(i).index() < index_max:
            self.congroups.append(Gamma(i))
            i += 1
        i = 1
        while Gamma0(i).index() < index_max:
            self.congroups.append(Gamma0(i))
            i += 1
        i = 2
        while Gamma1(i).index() < index_max:
            self.congroups.append(Gamma1(i))
            M = Zmod(i)
            U = [x for x in M if x.is_unit()]
            for j in range(1, len(U) - 1):
                self.congroups.append(GammaH(i, prandom.sample(U, j)))
            i += 1

        self.index = index

    def __repr__(self):
        r"""
        Return the string representation of ``self``.

        EXAMPLES::

            sage: from sage.modular.arithgroup.tests import Test
            sage: Test().__repr__()
            'Arithmetic subgroup testing class'
        """
        return "Arithmetic subgroup testing class"

    def _do(self, name):
        """
        Perform the test 'test_name', where name is specified as an
        argument. This function exists to avoid a call to eval.

        EXAMPLES::

            sage: from sage.modular.arithgroup.tests import Test
            sage: Test()._do("random")
            test_random
            ...
        """
        print("test_%s" % name)
        Test.__dict__["test_%s" % name](self)

    def random(self, seconds=0):
        """
        Perform random tests for a given number of seconds, or
        indefinitely if seconds is not specified.

        EXAMPLES::

            sage: from sage.modular.arithgroup.tests import Test
            sage: Test().random(1)
            test_random
            ...
        """
        self.test("random", seconds)

    def test(self, name, seconds=0):
        """
        Repeatedly run 'test_name', where name is passed as an
        argument. If seconds is nonzero, run for that many seconds. If
        seconds is 0, run indefinitely.

        EXAMPLES::

            sage: from sage.modular.arithgroup.tests import Test
            sage: T = Test()
            sage: T.test('relabel',seconds=1)
            test_relabel
            ...
            sage: T.test('congruence_groups',seconds=1)
            test_congruence_groups
            ...
            sage: T.test('contains',seconds=1)
            test_contains
            ...
            sage: T.test('todd_coxeter',seconds=1)
            test_todd_coxeter
            ...
        """
        seconds = float(seconds)
        total = cputime()
        n = 1
        while seconds == 0 or cputime(total) < seconds:
            s = "** test_dimension: number %s" % n
            if seconds > 0:
                s += " (will stop after about %s seconds)" % seconds
            t = cputime()
            self._do(name)
            print("\ttime=%s\telapsed=%s" % (cputime(t), cputime(total)))
            n += 1

    def test_random(self):
        """
        Do a random test from all the possible tests.

        EXAMPLES::

            sage: from sage.modular.arithgroup.tests import Test
            sage: Test().test_random() #random
            Doing random test
        """
        tests = [a for a in Test.__dict__
                 if a[:5] == "test_" and a != "test_random"]
        name = prandom.choice(tests)
        print("Doing random test %s" % name)
        Test.__dict__[name](self)

    def test_relabel(self):
        r"""
        Try the function canonical labels for a random even modular subgroup.

        EXAMPLES::

            sage: from sage.modular.arithgroup.tests import Test
            sage: Test().test_relabel() # random
        """
        if prandom.uniform(0, 1) < self.odd_probability:
            G = random_odd_arithgroup(self.index)
        else:
            G = random_even_arithgroup(self.index)

        G.relabel()
        s2 = G._S2
        s3 = G._S3
        l = G._L
        r = G._R

        # 0 should be stabilized by the mapping
        # used for renumbering so we start at 1
        p = list(range(1, self.index))

        for _ in range(10):
            prandom.shuffle(p)
            # we add 0 to the mapping
            pp = [0] + p
            ss2 = [None] * self.index
            ss3 = [None] * self.index
            ll = [None] * self.index
            rr = [None] * self.index
            for i in range(self.index):
                ss2[pp[i]] = pp[s2[i]]
                ss3[pp[i]] = pp[s3[i]]
                ll[pp[i]] = pp[l[i]]
                rr[pp[i]] = pp[r[i]]
            if G.is_even():
                GG = EvenArithmeticSubgroup_Permutation(ss2, ss3, ll, rr)
            else:
                GG = OddArithmeticSubgroup_Permutation(ss2, ss3, ll, rr)
            GG.relabel()

            for elt in ['_S2', '_S3', '_L', '_R']:
                if getattr(G, elt) != getattr(GG, elt):
                    print("s2 = %s" % str(s2))
                    print("s3 = %s" % str(s3))
                    print("ss2 = %s" % str(ss2))
                    print("ss3 = %s" % str(ss3))
                    print("pp = %s" % str(pp))
                    raise AssertionError("%s does not coincide" % elt)

    def test_congruence_groups(self):
        r"""
        Check whether the different implementations of methods for congruence
        groups and generic arithmetic group by permutations return the same
        results.

        EXAMPLES::

            sage: from sage.modular.arithgroup.tests import Test
            sage: Test().test_congruence_groups() #random
        """
        G = prandom.choice(self.congroups)
        GG = G.as_permutation_group()

        if not GG.is_congruence():
            raise AssertionError("Hsu congruence test failed")

        methods = [
            'index',
            'is_odd',
            'is_even',
            'is_normal',
            'ncusps',
            'nregcusps',
            'nirregcusps',
            'nu2',
            'nu3',
            'generalised_level']

        for f in methods:
            if getattr(G, f)() != getattr(GG, f)():
                raise AssertionError("results of %s does not coincide for %s" % (f, G))

        if sorted(G.cusp_width(c) for c in G.cusps()) != GG.cusp_widths():
            raise AssertionError("Cusps widths are different for %s" % G)

        for _ in range(20):
            m = GG.random_element()
            if m not in G:
                raise AssertionError("random element generated by perm. group not in %s" % str(G))

    def test_contains(self):
        r"""
        Test whether the random generator for arithgroup perms gives matrices in
        the group.

        EXAMPLES::

            sage: from sage.modular.arithgroup.tests import Test
            sage: Test().test_contains() #random
        """
        if prandom.uniform(0, 1) < self.odd_probability:
            G = random_odd_arithgroup(self.index)
        else:
            G = random_even_arithgroup(self.index)

        for _ in range(20):
            g = G.random_element()
            if G.random_element() not in G:
                raise AssertionError("%s not in %s" % (g, G))

    def test_spanning_trees(self):
        r"""
        Test coset representatives obtained from spanning trees for even
        subgroup (Kulkarni's method with generators ``S2``, ``S3`` and Verrill's
        method with generators ``L``, ``S2``).

        EXAMPLES::

            sage: from sage.modular.arithgroup.tests import Test
            sage: Test().test_spanning_trees() #random
        """
        from sage.misc.misc_c import prod

        from .all import SL2Z
        from .arithgroup_perm import Lm, S2m, S3m

        G = random_even_arithgroup(self.index)

        m = {'l': Lm, 's': S2m}
        tree, reps, wreps, gens = G._spanning_tree_verrill()
        assert reps[0] == SL2Z([1, 0, 0, 1])
        assert wreps[0] == ''
        for i in range(1, self.index):
            assert prod(m[letter] for letter in wreps[i]) == reps[i]
        tree, reps, wreps, gens = G._spanning_tree_verrill(on_right=False)
        assert reps[0] == SL2Z([1, 0, 0, 1])
        assert wreps[0] == ''
        for i in range(1, self.index):
            assert prod(m[letter] for letter in wreps[i]) == reps[i]

        m = {'s2': S2m, 's3': S3m}
        tree, reps, wreps, gens = G._spanning_tree_kulkarni()
        assert reps[0] == SL2Z([1, 0, 0, 1])
        assert wreps[0] == []
        for i in range(1, self.index):
            assert prod(m[letter] for letter in wreps[i]) == reps[i]
        tree, reps, wreps, gens = G._spanning_tree_kulkarni(on_right=False)
        assert reps[0] == SL2Z([1, 0, 0, 1])
        assert wreps[0] == []
        for i in range(1, self.index):
            assert prod(m[letter] for letter in wreps[i]) == reps[i]

    def test_todd_coxeter(self):
        r"""
        Test representatives of Todd-Coxeter algorithm.

        EXAMPLES::

            sage: from sage.modular.arithgroup.tests import Test
            sage: Test().test_todd_coxeter() #random
        """
        from .all import SL2Z
        from .arithgroup_perm import Lm, S2m, S3m

        G = random_even_arithgroup(self.index)

        reps, gens, l, s2 = G.todd_coxeter_l_s2()
        assert reps[0] == SL2Z([1, 0, 0, 1])
        assert len(reps) == G.index()
        for i in range(1, len(reps)):
            assert reps[i] not in G
            assert reps[i] * S2m * ~reps[s2[i]] in G
            assert reps[i] * Lm * ~reps[l[i]] in G
            for j in range(i + 1, len(reps)):
                assert reps[i] * ~reps[j] not in G
                assert reps[j] * ~reps[i] not in G

        reps, gens, s2, s3 = G.todd_coxeter_s2_s3()
        assert reps[0] == SL2Z([1, 0, 0, 1])
        assert len(reps) == G.index()
        for i in range(1, len(reps)):
            assert reps[i] not in G
            assert reps[i] * S2m * ~reps[s2[i]] in G
            assert reps[i] * S3m * ~reps[s3[i]] in G
            for j in range(i + 1, len(reps)):
                assert reps[i] * ~reps[j] not in G
                assert reps[j] * ~reps[i] not in G



from sage.misc.lazy_import import lazy_import

lazy_import('sage.groups.matrix_gps.linear', 'GL')
lazy_import('sage.groups.matrix_gps.linear', 'SL')
lazy_import('sage.groups.matrix_gps.symplectic', 'Sp')
lazy_import('sage.groups.matrix_gps.unitary', 'SU')
lazy_import('sage.groups.matrix_gps.unitary', 'GU')
lazy_import('sage.groups.matrix_gps.orthogonal', 'GO')
lazy_import('sage.groups.matrix_gps.orthogonal', 'SO')
lazy_import('sage.groups.matrix_gps.finitely_generated', 'MatrixGroup')
lazy_import('sage.groups.matrix_gps.finitely_generated', 'QuaternionMatrixGroupGF3')

import sage.groups.matrix_gps.pickling_overrides



# sage.doctest: needs numpy
"""
Complex Interpolation

AUTHORS:

- Ethan Van Andel (2009): initial version

Development supported by NSF award No. 0702939.
"""

# ****************************************************************************
#       Copyright (C) 2009 Ethan Van Andel <evlutte@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import numpy as np
cimport numpy as np

if int(np.version.short_version[0]) > 1:
    np.set_printoptions(legacy="1.25")

from math import pi
cdef double TWOPI = 2*pi


def polygon_spline(pts):
    """
    Create a polygon from a set of complex or `(x,y)` points. The polygon
    will be a parametric curve from 0 to 2*pi. The returned values will be
    complex, not `(x,y)`.

    INPUT:

    - ``pts`` -- list or array of complex numbers of tuples of the form `(x,y)`

    EXAMPLES:

    A simple square::

        sage: pts = [(-1, -1), (1, -1), (1, 1), (-1, 1)]
        sage: ps = polygon_spline(pts)
        sage: fx = lambda x: ps.value(x).real
        sage: fy = lambda x: ps.value(x).imag
        sage: show(parametric_plot((fx, fy), (0, 2*pi)))                                # needs sage.plot sage.symbolic
        sage: m = Riemann_Map([lambda x: ps.value(real(x))],
        ....:                 [lambda x: ps.derivative(real(x))], 0)
        sage: show(m.plot_colored() + m.plot_spiderweb())                               # needs sage.plot

    Polygon approximation of a circle::

        sage: # needs sage.symbolic
        sage: pts = [e^(I*t / 25) for t in range(25)]
        sage: ps = polygon_spline(pts)
        sage: ps.derivative(2)
        (-0.0470303661...+0.1520363883...j)
    """
    return PSpline(pts)


cdef class PSpline:
    """
    A ``CCSpline`` object contains a polygon interpolation of a figure
    in the complex plane.

    EXAMPLES:

    A simple ``square``::

        sage: pts = [(-1, -1), (1, -1), (1, 1), (-1, 1)]
        sage: ps = polygon_spline(pts)
        sage: ps.value(0)
        (-1-1j)
        sage: ps.derivative(0)
        (1.27323954...+0j)
    """
    cdef int N
    cdef np.ndarray pts

    def __init__(self, pts):
        """
        TESTS::

            sage: pts = [(-1, -1), (1, -1), (1, 1), (-1, 1)]
            sage: ps = polygon_spline(pts)
        """
        if isinstance(pts[0], tuple):
            self.pts = np.array(
                [complex(i[0], i[1]) for i in pts], dtype=np.complex128)
        else:
            self.pts = np.array(pts, dtype=np.complex128)
        self.N = len(pts)

    def value(self, double t):
        """
        Return the derivative (speed and direction of the curve) of a
        given point from the parameter ``t``.

        INPUT:

        - ``t`` -- double; the parameter value for the parameterized curve,
          between 0 and 2*pi

        OUTPUT:

        A complex number representing the point on the polygon corresponding
        to the input ``t``.

        EXAMPLES:

        ::

            sage: pts = [(-1, -1), (1, -1), (1, 1), (-1, 1)]
            sage: ps = polygon_spline(pts)
            sage: ps.value(.5)
            (-0.363380227632...-1j)
            sage: ps.value(0) - ps.value(2*RDF.pi())
            0j
            sage: ps.value(10)
            (0.26760455264...+1j)
        """
        cdef double t1 = ((t / TWOPI*self.N) % self.N)
        pt1 = self.pts[int(t1)]
        pt2 = self.pts[(int(t1) + 1) % self.N]
        return pt1 + (pt2 - pt1) * (t1 - int(t1))

    def derivative(self, double t):
        """
        Return the derivative (speed and direction of the curve) of a
        given point from the parameter ``t``.

        INPUT:

        - ``t`` -- double; the parameter value for the parameterized curve,
          between 0 and 2*pi

        OUTPUT:

        A complex number representing the derivative at the point on the
        polygon corresponding to the input ``t``.

        EXAMPLES:

        ::

            sage: pts = [(-1, -1), (1, -1), (1, 1), (-1, 1)]
            sage: ps = polygon_spline(pts)
            sage: ps.derivative(1 / 3)
            (1.27323954473...+0j)
            sage: from math import pi
            sage: ps.derivative(0) - ps.derivative(2*pi)
            0j
            sage: ps.derivative(10)
            (-1.27323954473...+0j)
        """
        cdef double t1 = ((t / TWOPI*self.N) % self.N)
        pt1 = self.pts[int(t1)]
        pt2 = self.pts[(int(t1) + 1) % self.N]
        return (pt2 - pt1) * self.N / TWOPI


def complex_cubic_spline(pts):
    """
    Create a cubic spline interpolated figure from a set of complex or
    `(x,y)` points. The figure will be a parametric curve from 0 to 2*pi.
    The returned values will be complex, not `(x,y)`.

    INPUT:

    - ``pts`` -- list or array of complex numbers, or tuples of the form `(x,y)`

    EXAMPLES:

    A simple ``square``::

        sage: pts = [(-1, -1), (1, -1), (1, 1), (-1, 1)]
        sage: cs = complex_cubic_spline(pts)
        sage: fx = lambda x: cs.value(x).real
        sage: fy = lambda x: cs.value(x).imag
        sage: from math import pi
        sage: show(parametric_plot((fx, fy), (0, 2*pi)))                                # needs sage.plot
        sage: m = Riemann_Map([lambda x: cs.value(real(x))],
        ....:                 [lambda x: cs.derivative(real(x))], 0)
        sage: show(m.plot_colored() + m.plot_spiderweb())                               # needs sage.plot

    Polygon approximation of a circle::

        sage: from cmath import exp
        sage: pts = [exp(1j * t / 25) for t in range(25)]
        sage: cs = complex_cubic_spline(pts)
        sage: cs.derivative(2)
        (-0.0497765406583...+0.151095006434...j)
    """
    return CCSpline(pts)


cdef class CCSpline:
    """
    A ``CCSpline`` object contains a cubic interpolation of a figure
    in the complex plane.

    EXAMPLES:

    A simple ``square``::

        sage: pts = [(-1, -1), (1, -1), (1, 1), (-1, 1)]
        sage: cs = complex_cubic_spline(pts)
        sage: cs.value(0)
        (-1-1j)
        sage: cs.derivative(0)
        (0.9549296...-0.9549296...j)
    """
    cdef int N
    cdef np.ndarray avec, bvec, cvec, dvec

    # standard cubic interpolation method
    def __init__(self, pts):
        """
        TESTS::

            sage: pts = [(-1, -1), (1, -1), (1, 1), (-1, 1)]
            sage: cs = complex_cubic_spline(pts)
        """
        if isinstance(pts[0], tuple):
            pts = np.array(
                [complex(pt[0], pt[1]) for pt in pts], dtype=np.complex128)
        cdef int N, i
        N = len(pts)
        yvec = np.zeros(N, dtype=np.complex128)
        for i in range(N):
            yvec[i] = 3 * (pts[(i - 1) % N] - 2*pts[i] + pts[(i + 1) % N])
        bmat = np.zeros([N, N], dtype=np.complex128)
        for i in range(N):
            bmat[i, i] = 4
            bmat[(i - 1) % N, i] = 1
            bmat[(i + 1) % N, i] = 1
        bvec = (np.linalg.solve(bmat, yvec))
        cvec = np.zeros(N, dtype=np.complex128)
        for i in range(N):
            cvec[i] = (pts[(i + 1) % N] - pts[i] - 1.0/3.0 *
                       bvec[(i + 1) % N] - 2./3. * bvec[i])
        dvec = np.array(pts, dtype=np.complex128)
        avec = np.zeros(N, dtype=np.complex128)
        for i in range(N):
            avec[i] = 1.0/3.0 * (bvec[(i + 1) % N] - bvec[i])
        self.avec = avec
        self.bvec = bvec
        self.cvec = cvec
        self.dvec = dvec
        self.N = N

    def value(self, double t):
        """
        Return the location of a given point from the parameter ``t``.

        INPUT:

        - ``t`` -- double; the parameter value for the parameterized curve,
          between 0 and 2*pi

        OUTPUT:

        A complex number representing the point on the figure corresponding
        to the input ``t``.

        EXAMPLES:

        ::

            sage: pts = [(-1, -1), (1, -1), (1, 1), (-1, 1)]
            sage: cs = complex_cubic_spline(pts)
            sage: cs.value(4 / 7)
            (-0.303961332787...-1.34716728183...j)
            sage: from math import pi
            sage: cs.value(0) - cs.value(2*pi)
            0j
            sage: cs.value(-2.73452)
            (0.934561222231...+0.881366116402...j)
        """
        cdef double t1 = (t / TWOPI * self.N) % self.N
        cdef int j = int(t1)
        return (self.avec[j] * (t1 - j)**3 + self.bvec[j] * (t1 - j)**2 +
                self.cvec[j] * (t1 - j) + self.dvec[j])

    def derivative(self, double t):
        """
        Return the derivative (speed and direction of the curve) of a
        given point from the parameter ``t``.

        INPUT:

        - ``t`` -- double; the parameter value for the parameterized curve,
          between 0 and 2*pi

        OUTPUT:

        A complex number representing the derivative at the point on the
        figure corresponding to the input ``t``.

        EXAMPLES::

            sage: pts = [(-1, -1), (1, -1), (1, 1), (-1, 1)]
            sage: cs = complex_cubic_spline(pts)
            sage: cs.derivative(3 / 5)
            (1.40578892327...-0.225417136326...j)
            sage: from math import pi
            sage: cs.derivative(0) - cs.derivative(2 * pi)
            0j
            sage: cs.derivative(-6)
            (2.52047692949...-1.89392588310...j)
        """
        cdef double t1 = (t / TWOPI * self.N) % self.N
        cdef int j = int(t1)
        return (3 * self.avec[j] * (t1 - j)**2 + 2 * self.bvec[j] * (t1 - j) +
                self.cvec[j]) / TWOPI * self.N



r"""
Concrete classes related to modules with a distinguished basis

This module provides concrete classes for various constructions
related to modules with a distinguished basis:

- :mod:`morphism` -- Concrete classes for morphisms of modules with basis

.. SEEALSO:: The category :class:`ModulesWithBasis`
"""
# install the docstring of this module to the containing package
from sage.misc.namespace_package import install_doc
install_doc(__package__, __doc__)
del install_doc
